import os

from PySide6.QtCore import Qt, QAbstractItemModel, QModelIndex, Signal, QPoint, QUrl
from PySide6.QtGui import QCursor, QColor, QIcon, QDesktopServices, QShortcut, QKeySequence
from PySide6.QtWidgets import QWidget, QVBoxLayout
from qfluentwidgets import TreeItemDelegate, TreeView, RoundMenu, Action, MessageBox

from NepTrainKit.core import MessageManager
from NepTrainKit.core.dataset import DatasetManager
from NepTrainKit.core.dataset.services import ProjectItem, ModelItem
from NepTrainKit.core.types import ModelTypeIcon
from NepTrainKit.custom_widget import TreeModel, TreeItem,TagDelegate
from NepTrainKit.custom_widget.dialog import ModelInfoMessageBox, AdvancedModelSearchDialog, TagManageDialog


# from NepTrainKit.custom_widget import DatasetItemModel







class ModelItemWidget(QWidget,DatasetManager):
    model_item_dict={}
    projectChangedSignal=Signal(int)
    def __init__(self, parent=None):
        super(ModelItemWidget, self).__init__(parent)
        self._parent=parent
        self.project:ProjectItem
        # self.project_service.create_project("测试","222")
        #
        # self.project_service.create_project("测试1","222")
        self._view = TreeView()


        self._view.clicked.connect(self.item_clicked)
        self._view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

        # self._view.setIndentation(10)
        self._view.header().setDefaultSectionSize(50)



        self._view.header().setStretchLastSection(True)
        self._model=TreeModel()

        self._view.setModel(self._model)
        self._model.setHeader(["ID","Name","Size","E(meV/atom)","F(meV/Å)","V(meV/atom)", "Tags","Create At"])
        self._view.setItemDelegateForColumn(6, TagDelegate(self._model))
        width=[90,200,60,90,80,90,200,30]
        for col,w in enumerate(width):
            self._view.setColumnWidth(col,w)


        # self._view.setColumnHidden(1,True)
        # self._view.setColumnWidth(0,110)

        self.create_menu()
        self._layout = QVBoxLayout(self)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0,0,0,0)
        self._layout.addWidget(self._view)

        self.search_shortcut = QShortcut(QKeySequence("Ctrl+F"),self,context=Qt.ShortcutContext.WindowShortcut)
        self.search_shortcut.activated.connect(self.show_search_dialog)

    def item_clicked(self,index):
        item = index.internalPointer()

        self.projectChangedSignal.emit(item.data(1))
    def create_menu(self):
        self._menu_pos = QPoint()
        self.menu = RoundMenu(parent=self)

        create_action=Action("New",self.menu)
        create_action.triggered.connect(lambda :self.create_model(modify=False))
        self.menu.addAction(create_action)
        modify_action=Action("Modify",self.menu)
        modify_action.triggered.connect(lambda :self.create_model(modify=True))
        self.menu.addAction(modify_action)

        open_action = Action("Open Folder", self.menu)
        open_action.triggered.connect(self.open_folder)
        self.menu.addAction(open_action)

        delete_action = Action("Delete", self.menu)
        delete_action.triggered.connect(self.remove_model)
        self.menu.addAction(delete_action)
        tag_action = Action("Manage Tags", self.menu)
        tag_action.triggered.connect(self.manage_tags)
        self.menu.addAction(tag_action)
        self._view.customContextMenuRequested.connect(self.show_menu)
    def show_menu(self,pos):
        self._menu_pos=pos
        self.menu.exec_(self.mapToGlobal(pos))
    def manage_tags(self):
        dlg = TagManageDialog(self.tag_service, self._parent)
        dlg.exec_()
        # if dlg.tag_changed:
        #     self.

    def _build_tree(self,model,parent:TreeItem):
        child = TreeItem((model.model_id,
                              model.name,
                              model.data_size,
                              model.energy,
                              model.force ,
                              model.virial ,
                          [{"name":tag.name,"color":tag.color} for tag in model.tags],

                              model.created_at.strftime("%Y-%m-%d %H:%M:%S"),))

        child.icon = QIcon(ModelTypeIcon.NEP)

        self.model_item_dict[model.model_id] = model
        parent.appendChild(child)
        for item in model.children:
            self._build_tree(item,child)
        return child
    def load_models_by_project(self, project):
        self._model.clear()
        self.project=project
        models=self.model_service.get_models_by_project_id(project.project_id)
        self.add_models_to_table(models)
    def add_models_to_table(self,models:list[ModelItem]):
        self._model.beginResetModel()
        for model in models:
            self._build_tree(model,self._model.rootItem)
            continue
        self._model.endResetModel()
    def create_model(self,modify=False):
        box = ModelInfoMessageBox(self._parent)
        index=self._view.indexAt(self._menu_pos)
        box.parent_combox.addItem("Top Model",userData=None)
        for  model in self.model_item_dict.values():
            box.parent_combox.addItem(f"{model.model_id}-{model.name}", userData=model.model_id)

        if index.row()!=-1:
            item=index.internalPointer()
            box.parent_combox.setCurrentText(f"{item.data(0)}-{item.data(1)}")
            model_id=item.data(0)
        else:
            box.parent_combox.setCurrentText("Top Model")
            if modify:
                return
            model_id=None

        if modify:
            current_model = self.model_item_dict[model_id]
            box.model_name_edit.setText(current_model.name)
            box.model_note_edit.setText(current_model.notes)
            box.train_path_edit.setText(current_model.model_path)
            box.model_type_combox.setText(current_model.model_type)
            box.energy_spinBox.setText(str(current_model.energy))
            box.force_spinBox.setText(str(current_model.force))
            box.virial_spinBox.setText(str(current_model.virial))
            for tag in current_model.tags:
                box.add_tag(tag.name)


            if current_model.parent_id is not None:
                parent_model = self.model_item_dict[current_model.parent_id]
                box.parent_combox.setCurrentText(f"{parent_model.model_id}-{parent_model.name}")
            else:
                box.parent_combox.setCurrentText("Top Model")

        if not box.exec_():
            return
        data = box.get_dict()
        data["project_id"]=self.project.project_id

        if modify:
            self.model_service.modify_model(current_model.model_id,
                                                **data)
            self.load_models_by_project(self.project)

            MessageManager.send_success_message("Model modification successful")
            return

        project = self.model_service.add_version(
            **data

        )
        if project is None:
            MessageManager.send_error_message("Failed to create model")
        else:
            MessageManager.send_success_message("Model created successfully")
            self.load_models_by_project(self.project)
    def remove_model(self):
        index = self._view.indexAt(self._menu_pos)

        if index.row() == -1:
            return

        item = index.internalPointer()

        model_id = item.data(0)
        box = MessageBox("Ask",
                         "Do you want to delete this item?\nIf you delete it, all items under it will be deleted!",
                         self._parent)
        box.exec_()
        if box.result() == 0:
            return

        self.model_service.remove_model(model_id=model_id)

        MessageManager.send_success_message("Model deleted successfully")
        self.load_models_by_project(self.project)

    def open_folder(self):
        index = self._view.indexAt(self._menu_pos)

        if index.row() == -1:
            return

        item = index.internalPointer()
        model_id = item.data(0)
        model=self.model_item_dict[model_id]
        path=model.model_path
        if path.startswith("http"):
            QDesktopServices.openUrl(QUrl(path))
        else:
            if os.path.exists(path):
                QDesktopServices.openUrl(QUrl("file:///"+path))
    def on_search(self,params):
        models= self.model_service.search_models_advanced(**params)
        self._model.clear()
        self.add_models_to_table(models)

    def show_search_dialog(self):

        box = AdvancedModelSearchDialog(self._parent)
        box.searchRequested.connect(self.on_search)
        box.show()


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/12/17 12:57
# @Author  : 兵
# @email    : 1747193328@qq.com
from typing import Union

from PySide6.QtCore import Signal
from PySide6.QtGui import QIcon, Qt
from qfluentwidgets import OptionsConfigItem, FluentIconBase, ComboBox, SettingCard, DoubleSpinBox, LineEdit
from PySide6.QtWidgets import QPushButton
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QColorDialog


class MyComboBoxSettingCard(SettingCard):
    """ Setting card with a combo box """
    optionChanged = Signal(str)

    def __init__(self, configItem: OptionsConfigItem,
                 icon: Union[str, QIcon, FluentIconBase],
                 title, content , texts:list[str], default=None,
                 parent=None):
        """
        Parameters
        ----------
        configItem: OptionsConfigItem
            configuration item operated by the card

        icon: str | QIcon | FluentIconBase
            the icon to be drawn

        title: str
            the title of card

        content: str
            the content of card

        texts: List[str]
            the text of items

        parent: QWidget
            parent widget
        """
        super().__init__(icon, title, content, parent)
        self.configItem = configItem
        self.comboBox = ComboBox(self)
        self.hBoxLayout.addWidget(self.comboBox, 0, Qt.AlignmentFlag.AlignRight)
        self.hBoxLayout.addSpacing(16)

        self.optionToText = {o: t for o, t in zip(configItem.options, texts)}
        for text, option in zip(texts, configItem.options):
            self.comboBox.addItem(text, userData=option)
        if default is not None:

            self.comboBox.setCurrentText(default)
        self.comboBox.currentTextChanged.connect(self.optionChanged)
        configItem.valueChanged.connect(self.setValue)



    def setValue(self, value):
        if value not in self.optionToText:
            return

        self.comboBox.setCurrentText(self.optionToText[value])





class DoubleSpinBoxSettingCard(SettingCard):
    """ Setting card with a push button """

    valueChanged = Signal(float)

    def __init__(self,   icon: Union[str, QIcon, FluentIconBase], title, content=None, parent=None):
        """
        Parameters
        ----------
        text: str
            the text of push button

        icon: str | QIcon | FluentIconBase
            the icon to be drawn

        title: str
            the title of card

        content: str
            the content of card

        parent: QWidget
            parent widget
        """
        super().__init__(icon, title, content, parent)
        self.doubleSpinBox = DoubleSpinBox(  self)
        self.doubleSpinBox.setDecimals(2)
        self.doubleSpinBox.setSingleStep(0.1)
        self.doubleSpinBox.setMinimumWidth(200)
        self.hBoxLayout.addWidget(self.doubleSpinBox, 0, Qt.AlignmentFlag.AlignRight)
        self.hBoxLayout.addSpacing(16)
        self.doubleSpinBox.valueChanged.connect(self.valueChanged)
    def setValue(self, value):
        self.doubleSpinBox.setValue(value)
    def setRange(self, min, max):
        self.doubleSpinBox.setRange(min, max)


class ColorSettingCard(SettingCard):
    """Setting card with a color picker button"""

    colorChanged = Signal(str)

    def __init__(self, icon: Union[str, QIcon, FluentIconBase], title: str, content: str | None = None,
                 parent=None):
        super().__init__(icon, title, content, parent)
        self._color = QColor("#000000")
        self.button = QPushButton(self)
        self.button.setFixedSize(64, 24)
        self.button.clicked.connect(self._choose_color)
        self.hBoxLayout.addWidget(self.button, 0, Qt.AlignmentFlag.AlignRight)
        self.hBoxLayout.addSpacing(16)
        self._apply_button_style()

    def _apply_button_style(self):
        # show current color as button background
        self.button.setStyleSheet(
            f"QPushButton {{ background-color: {self._color.name()}; border: 1px solid #999; border-radius: 4px; }}"
        )

    def _choose_color(self):
        col = QColorDialog.getColor(self._color, self)
        if col.isValid():
            self._color = col
            self._apply_button_style()
            # emit as hex RGB string
            self.colorChanged.emit(self._color.name())

    def setValue(self, value: str):
        try:
            col = QColor(value)
            if col.isValid():
                self._color = col
                self._apply_button_style()
        except Exception:
            pass


class LineEditSettingCard(SettingCard):
    """Setting card with a single-line text input"""

    textChanged = Signal(str)

    def __init__(self, icon: Union[str, QIcon, FluentIconBase], title: str, content: str | None = None,
                 parent=None):
        super().__init__(icon, title, content, parent)
        self.lineEdit = LineEdit(self)
        self.lineEdit.setMinimumWidth(220)
        self.hBoxLayout.addWidget(self.lineEdit, 0, Qt.AlignmentFlag.AlignRight)
        self.hBoxLayout.addSpacing(16)
        self.lineEdit.textChanged.connect(self.textChanged)

    def setValue(self, text: str):
        self.lineEdit.setText(str(text) if text is not None else "")

    def value(self) -> str:
        return self.lineEdit.text()



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/12/2 20:02
# @Author  : 兵
# @email    : 1747193328@qq.com
import sys
from enum import Enum

from PySide6.QtCore import Qt
from PySide6.QtGui import QBrush, QColor, QPen, QIcon
from NepTrainKit.config import Config

if sys.version_info >= (3, 11):
    from enum import StrEnum          # 3.11+
else:
    from enum import Enum
    class StrEnum(str, Enum):         # 3.10- 的回退
        pass
#pyqtgragh导入很慢  所以先拷贝过来 后面看要不要优化掉
def mkPen(*args, **kwargs):
    """
    Convenience function for constructing QPen.

    Examples::

        mkPen(color)
        mkPen(color, width=2)
        mkPen(cosmetic=False, width=4.5, color='r')
        mkPen({'color': "#FF0", width: 2})
        mkPen(None)   # (no pen)

    In these examples, *color* may be replaced with any arguments accepted by :func:`mkColor() <pyqtgraph.mkColor>`    """
    color = kwargs.get('color', None)
    width = kwargs.get('width', 1)
    style = kwargs.get('style', None)
    dash = kwargs.get('dash', None)
    cosmetic = kwargs.get('cosmetic', True)
    hsv = kwargs.get('hsv', None)

    if len(args) == 1:
        arg = args[0]
        if isinstance(arg, dict):
            return mkPen(**arg)
        if isinstance(arg, QPen):
            return QPen(arg)  ## return a copy of this pen
        elif arg is None:
            style = Qt.PenStyle.NoPen
        else:
            color = arg
    if len(args) > 1:
        color = args


    color = QColor(color)

    pen = QPen(QBrush(color), width)
    pen.setCosmetic(cosmetic)
    if style is not None:
        pen.setStyle(style)
    if dash is not None:
        pen.setDashPattern(dash)

    # for width > 1.0, we are drawing many short segments to emulate a
    # single polyline. the default SquareCap style causes artifacts.
    # these artifacts can be avoided by using RoundCap.
    # this does have a performance penalty, so enable it only
    # for thicker line widths where the artifacts are visible.
    if width > 4.0:
        pen.setCapStyle(Qt.PenCapStyle.RoundCap)

    return pen

class ForcesMode(StrEnum):
    Raw="Raw"
    Norm="Norm"

class CanvasMode(StrEnum):
    VISPY= "vispy"
    PYQTGRAPH = "pyqtgraph"

class SearchType(StrEnum):
    TAG="Config_type"
    FORMULA="formula"

class NepBackend(StrEnum):
    AUTO = "auto"
    GPU = "gpu"
    CPU = "cpu"

class Base:
    @classmethod
    def get(cls,name):
        if hasattr(cls, name):
            return getattr(cls, name)
        else:
            return getattr(cls,"Default")

def _get_color(section: str, option: str, default_hex: str) -> QColor:
    val = Config.get(section, option, default_hex)
    try:
        c = QColor(val)
        if c.isValid():
            return c
        return QColor(default_hex)
    except Exception:
        return QColor(default_hex)


class Pens(Base):
    # Initialize from config with sensible defaults
    @classmethod
    def update_from_config(cls):
        edge = _get_color("plot", "marker_edge_color", "#07519C")
        current = _get_color("plot", "current_color", "#FF0000")
        line = _get_color("plot", "line_color", "#FF0000")

        cls.Default = mkPen(color=edge, width=0.8)
        cls.Energy = cls.Default
        cls.Force = cls.Default
        cls.Virial = cls.Default
        cls.Stress = cls.Default
        cls.Descriptor = cls.Default
        cls.Current = mkPen(color=current, width=1)
        cls.Line = mkPen(color=line, width=2)

    def __getattr__(self, item):
        return getattr(self.Default, item)

class Brushes(Base):
    # Initialize from config
    @classmethod
    def update_from_config(cls):
        face = _get_color("plot", "marker_face_color", "#FFFFFF")
        # optional alpha channel from separate setting
        alpha = Config.getint("plot", "marker_face_alpha", 0) or 0
        face.setAlpha(int(max(0, min(255, alpha))))

        show = _get_color("plot", "show_color", "#00FF00")
        selected = _get_color("plot", "selected_color", "#FF0000")
        current = _get_color("plot", "current_color", "#FF0000")

        cls.BlueBrush = QBrush(QColor(0, 0, 255))
        cls.YellowBrush = QBrush(QColor(255, 255, 0))
        cls.Default = QBrush(face)
        cls.Energy = cls.Default
        cls.Force = cls.Default
        cls.Virial = cls.Default
        cls.Stress = cls.Default
        cls.Descriptor = cls.Default
        cls.Show = QBrush(show)
        cls.Selected = QBrush(selected)
        cls.Current = QBrush(current)

    def __getattr__(self, item):
        return getattr(self.Default, item)

class ModelTypeIcon(Base):

    NEP=':/images/src/images/gpumd_new.png'

# Initialize pens/brushes on import
Pens.update_from_config()
Brushes.update_from_config()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import traceback
from pathlib import Path
import os
import numpy as np
import numpy.typing as npt
from PySide6.QtCore import QObject, Signal
from loguru import logger
import glob
from .base import NepPlotData, StructureData, ResultData,DPPlotData
from NepTrainKit.core.structure import Structure, load_npy_structure,save_npy_structure
from .utils import parse_array_by_atomnum,read_nep_out_file
from NepTrainKit.config import Config

from .. import   MessageManager
from ..calculator import run_nep_calculator
from ..types import NepBackend
from ... import module_path
def is_deepmd_path(folder)-> bool:
    if os.path.exists(os.path.join(folder,"type.raw")):
        return True
    if glob.glob(os.path.join(folder, "**/type.raw"), recursive=True):
        return True
    return False
class DeepmdResultData(ResultData):
    _energy_dataset:DPPlotData
    _force_dataset:DPPlotData
    _spin_dataset:DPPlotData
    _virial_dataset:DPPlotData
    def __init__(self, nep_txt_path: Path|str,
                 data_xyz_path: Path|str,
                 energy_out_path: Path|str,
                 force_out_path: Path|str,
                 virial_out_path: Path|str,
                 descriptor_path: Path|str,
                 spin_out_path: Path|str|None=None
                 ):
        super().__init__(nep_txt_path, data_xyz_path,descriptor_path)
        self.energy_out_path = Path(energy_out_path)
        self.force_out_path = Path(force_out_path)
        self.spin_out_path = Path(spin_out_path) if spin_out_path is not None else None
        self.virial_out_path = Path(virial_out_path)



    @classmethod
    def from_path(cls, path, *, structures: list[Structure] | None = None):
        dataset_path = Path(path)

        # file_name=dataset_path.name
        nep_txt_path = dataset_path.with_name(f"nep.txt")
        if not nep_txt_path.exists():
            nep89_path = os.path.join(module_path, "Config/nep89.txt")
            nep_txt_path=Path(nep89_path)
        descriptor_path = dataset_path.with_name(f"descriptor.out")
        e_path = list(dataset_path.parent.glob("*.e_peratom.out") )
        if e_path:
            e_path = e_path[0]
            suffix = (e_path.name.replace(".e_peratom.out",""))

        else:
            suffix="detail"



        energy_out_path = dataset_path.with_name(f"{suffix}.e_peratom.out")
        force_out_path = dataset_path.with_name(f"{suffix}.fr.out")
        if  not force_out_path.exists():
            force_out_path = dataset_path.with_name(f"{suffix}.f.out")

        # stress_out_path = dataset_path.with_name(f"{suffix}.v.out")
        virial_out_path = dataset_path.with_name(f"{suffix}.v_peratom.out")
        spin_out_path=  dataset_path.with_name(f"{suffix}.fm.out")
        if not spin_out_path.exists():
            spin_out_path = None
        inst = cls(nep_txt_path,dataset_path,energy_out_path,force_out_path,virial_out_path,descriptor_path,spin_out_path=spin_out_path)
        # DeepMD loader ignores in-memory structures; it reads its own format.
        return inst

    def load_structures(self):
        """
        加载训练集的结构
        :return:
        """
        structures = load_npy_structure(self.data_xyz_path, cancel_event=self.cancel_event)
        self._atoms_dataset = StructureData(structures)
        self.atoms_num_list = np.array([len(s) for s in structures])





    @property
    def datasets(self):
        if self.spin_out_path is None:
            return [self.energy, self.force,  self.virial, self.descriptor]
        else:
            return [self.energy, self.force,self.spin, self.virial, self.descriptor]
    @property
    def energy(self):
        return self._energy_dataset

    @property
    def force(self):
        return self._force_dataset

    @property
    def spin(self):
        return self._spin_dataset

    @property
    def virial(self):
        return self._virial_dataset


    def _load_dataset(self) -> None:

        if self._should_recalculate( ):
            energy_array, force_array, virial_array = self._recalculate_and_save( )
        else:
            energy_array=read_nep_out_file(self.energy_out_path,ndmin=2)
            force_array=read_nep_out_file(self.force_out_path,ndmin=2)
            virial_array=read_nep_out_file(self.virial_out_path,ndmin=2)
            if energy_array.shape[0]!=self.atoms_num_list.shape[0]:
                self.energy_out_path.unlink(True)
                self.force_out_path.unlink(True)
                if self.spin_out_path is not None:
                    self.spin_out_path.unlink(True)

                self.virial_out_path.unlink(True)


                return self._load_dataset()

        self._energy_dataset = DPPlotData(energy_array, title="energy")
        default_forces = Config.get("widget", "forces_data", "Row")
        if force_array.size != 0 and default_forces == "Norm":

            force_array = parse_array_by_atomnum(force_array, self.atoms_num_list, map_func=np.linalg.norm, axis=0)

            self._force_dataset = DPPlotData(force_array, title="force")
        else:
            self._force_dataset = DPPlotData(force_array, group_list=self.atoms_num_list, title="force")

        if self.spin_out_path is not None:
            spin_array=read_nep_out_file(self.spin_out_path,ndmin=2)
            group_list=[s.spin_num for s in self.structure.now_data]
            if (np.sum(group_list))!=0:
                self._spin_dataset = DPPlotData(spin_array,  group_list=group_list, title="spin")
            else:
                self.spin_out_path = None

        self._virial_dataset = DPPlotData(virial_array, title="virial")


    def _should_recalculate(self  ) -> bool:
        """判断是否需要重新计算 性质 数据。"""
        output_files_exist = any([
            self.energy_out_path.exists(),
            self.force_out_path.exists(),

            self.virial_out_path.exists()
        ])
        return   not output_files_exist

    def _save_energy_data(self, potentials: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:

        """保存能量数据到文件。"""

        try:
            ref_energies = np.array([s.per_atom_energy for s in self.structure.now_data], dtype=np.float32)

            if potentials.size  == 0:
                #计算失败 空数组
                energy_array = np.column_stack([ref_energies, ref_energies])
            else:
                energy_array = np.column_stack([ref_energies,potentials / self.atoms_num_list  ])
        except Exception:
            # logger.debug(traceback.format_exc())
            if potentials.size == 0:
                # 计算失败 空数组
                energy_array = np.column_stack([potentials, potentials])
            else:
                energy_array = np.column_stack([potentials / self.atoms_num_list, potentials / self.atoms_num_list])
        energy_array = energy_array.astype(np.float32)
        if energy_array.size != 0:
            np.savetxt(self.energy_out_path, energy_array, fmt='%10.8f')
        return energy_array

    def _save_force_data(self, forces: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """保存力数据到文件。"""
        try:
            ref_forces = np.vstack([s.forces for s in self.structure.now_data], dtype=np.float32)

            if forces.size == 0:
                # 计算失败 空数组
                forces_array = np.column_stack([ref_forces, ref_forces])

            else:
                forces_array = np.column_stack([ref_forces,forces ])
        except KeyError:
            MessageManager.send_warning_message("use nep3 calculator to calculate forces replace the original forces")
            forces_array = np.column_stack([forces, forces])

        except Exception:
            # logger.debug(traceback.format_exc())
            forces_array = np.column_stack([forces, forces])
            MessageManager.send_error_message("an error occurred while calculating forces. Please check the input file.")
        if forces_array.size != 0:
            np.savetxt(self.force_out_path, forces_array, fmt='%10.8f')


        return forces_array



    def _save_virial_and_data(self, virials: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """保存维里和应力数据到文件。"""

        try:
            ref_virials = np.vstack([s.nep_virial for s in self.structure.now_data], dtype=np.float32)


            if virials.size == 0:
                # 计算失败 空数组
                virials_array = np.column_stack([ref_virials, ref_virials])
            else:
                virials_array = np.column_stack([ref_virials,virials  ])
        except AttributeError:
            MessageManager.send_warning_message("use nep3 calculator to calculate virial replace the original virial")
            virials_array = np.column_stack([virials, virials])

        except Exception:
            MessageManager.send_error_message(f"An error occurred while calculating virial and stress. Please check the input file.")
            # logger.debug(traceback.format_exc())
            virials_array = np.column_stack([virials, virials])

        if virials_array.size != 0:
            np.savetxt(self.virial_out_path, virials_array, fmt='%10.8f')



        return virials_array

    def _recalculate_and_save(self ):

        try:
            nep_potentials_array, nep_forces_array, nep_virials_array=   self.nep_calc.calculate(self.structure.now_data.tolist())

            # nep_potentials_array, nep_forces_array, nep_virials_array=self.nep_calc_thread.func_result
            # nep_potentials_array, nep_forces_array, nep_virials_array = run_nep3_calculator_process(
            #     self.nep_txt_path.as_posix(),
            #     self.structure.now_data,"calculate")
            if nep_potentials_array.size == 0:
                MessageManager.send_warning_message("The nep calculator fails to calculate the potentials, use the original potentials instead.")


            energy_array = self._save_energy_data(nep_potentials_array)
            force_array = self._save_force_data(nep_forces_array)
            virial_array = self._save_virial_and_data(nep_virials_array[:, [0, 4, 8, 1, 5, 6]])
            return energy_array,force_array,virial_array
        except Exception as e:
            # logger.debug(traceback.format_exc())
            MessageManager.send_error_message(f"An error occurred while running NEP3 calculator: {e}")
            return np.array([]), np.array([]), np.array([])

    def export_model_xyz(self,save_path):
        """
        导出当前结构
        :param save_path: 保存路径，传入的是一个文件夹路径
        被删除的导出到export_remove_model
        被保留的导出到export_good_model
        """
        try:
            save_npy_structure(os.path.join(save_path, "export_good_model"),self.structure.now_data)
            save_npy_structure(os.path.join(save_path, "export_remove_model"),self.structure.remove_data)


            MessageManager.send_info_message(f"File exported to: {save_path}")
        except:
            MessageManager.send_info_message(f"An unknown error occurred while saving. The error message has been output to the log!")
            logger.error(traceback.format_exc())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import sys
import math
import random
from dataclasses import dataclass
from typing import Iterable, List, Optional, Sequence, Tuple

import numpy as np
from collections import deque, defaultdict


# Covalent radii (Å) — copied from the original script
COVALENT_RADII = {
    'H': 0.31, 'He': 0.28, 'Li': 1.28, 'Be': 0.96, 'B': 0.84, 'C': 0.76, 'N': 0.71, 'O': 0.66, 'F': 0.57, 'Ne': 0.58,
    'Na': 1.66, 'Mg': 1.41, 'Al': 1.21, 'Si': 1.11, 'P': 1.07, 'S': 1.05, 'Cl': 1.02, 'Ar': 1.06,
    'K': 2.03, 'Ca': 1.76, 'Sc': 1.70, 'Ti': 1.60, 'V': 1.53, 'Cr': 1.39, 'Mn': 1.39, 'Fe': 1.32, 'Co': 1.26, 'Ni': 1.24,
    'Cu': 1.32, 'Zn': 1.22, 'Ga': 1.22, 'Ge': 1.20, 'As': 1.19, 'Se': 1.20, 'Br': 1.20, 'Kr': 1.16,
    'Rb': 2.20, 'Sr': 1.95, 'Y': 1.90, 'Zr': 1.75, 'Nb': 1.64, 'Mo': 1.54, 'Tc': 1.47, 'Ru': 1.46, 'Rh': 1.42, 'Pd': 1.39,
    'Ag': 1.45, 'Cd': 1.44, 'In': 1.42, 'Sn': 1.39, 'Sb': 1.39, 'Te': 1.38, 'I': 1.39, 'Xe': 1.40,
    'Cs': 2.44, 'Ba': 2.15, 'La': 2.07, 'Ce': 2.04, 'Pr': 2.03, 'Nd': 2.01, 'Pm': 1.99, 'Sm': 1.98, 'Eu': 1.98, 'Gd': 1.96,
    'Tb': 1.94, 'Dy': 1.92, 'Ho': 1.92, 'Er': 1.89, 'Tm': 1.90, 'Yb': 1.87, 'Lu': 1.87,
    'Hf': 1.75, 'Ta': 1.70, 'W': 1.62, 'Re': 1.51, 'Os': 1.44, 'Ir': 1.41, 'Pt': 1.36, 'Au': 1.36, 'Hg': 1.32,
    'Tl': 1.45, 'Pb': 1.46, 'Bi': 1.48, 'Po': 1.40, 'At': 1.50, 'Rn': 1.50,
    'Fr': 2.60, 'Ra': 2.21, 'Ac': 2.15, 'Th': 2.06, 'Pa': 2.00, 'U': 1.96, 'Np': 1.90, 'Pu': 1.87, 'Am': 1.80, 'Cm': 1.69
}


@dataclass
class TorsionGuardParams:
    perturb_per_frame: int = 100
    torsion_range_deg: Tuple[float, float] = (-180.0, 180.0)
    max_torsions_per_conf: int = 5
    gaussian_sigma: float = 0.03
    pbc_mode: str = "auto"  # "auto" | "yes" | "no"

    local_mode_cutoff_atoms: int = 150
    local_torsion_max_subtree: int = 40

    bond_detect_factor: float = 1.15
    bond_keep_min_factor: float = 0.60
    bond_keep_max_factor: Optional[float] = None
    nonbond_min_factor: float = 0.80

    max_retries_per_frame: int = 12
    mult_bond_factor: float = 0.87
    nonpbc_box_size: float = 100.0
    # Pauling bond order parameters
    bo_c_const: float = 0.3
    bo_threshold: float = 0.2


# ---------- Geometry and PBC helpers ----------

def rotate_coords(coords: np.ndarray, atom_indices: Iterable[int], axis_point1: np.ndarray,
                  axis_point2: np.ndarray, angle_deg: float) -> np.ndarray:
    axis = np.array(axis_point2) - np.array(axis_point1)
    norm = np.linalg.norm(axis)
    if norm < 1e-12:
        return coords
    axis = axis / norm
    angle = np.deg2rad(angle_deg)
    cos_a, sin_a = np.cos(angle), np.sin(angle)
    ux, uy, uz = axis
    R = np.array([
        [cos_a + ux * ux * (1 - cos_a), ux * uy * (1 - cos_a) - uz * sin_a, ux * uz * (1 - cos_a) + uy * sin_a],
        [uy * ux * (1 - cos_a) + uz * sin_a, cos_a + uy * uy * (1 - cos_a), uy * uz * (1 - cos_a) - ux * sin_a],
        [uz * ux * (1 - cos_a) - uy * sin_a, uz * uy * (1 - cos_a) + ux * sin_a, cos_a + uz * uz * (1 - cos_a)]
    ])
    idx = list(atom_indices)
    shifted = coords[idx] - axis_point1
    coords[idx] = shifted.dot(R.T) + axis_point1
    return coords


def center_in_box(coords: np.ndarray, box_size: float) -> np.ndarray:
    min_xyz = coords.min(axis=0)
    max_xyz = coords.max(axis=0)
    center = (min_xyz + max_xyz) / 2.0
    box_center = np.array([box_size / 2.0] * 3)
    shift = box_center - center
    return coords + shift


def wrap_to_cell(coords: np.ndarray, cell: np.ndarray, inv_cell_T: np.ndarray) -> np.ndarray:
    frac = coords @ inv_cell_T
    frac = frac - np.floor(frac)
    return frac @ cell


def mic_delta(delta: np.ndarray, cell: np.ndarray, inv_cell_T: np.ndarray) -> np.ndarray:
    dfrac = delta @ inv_cell_T
    dfrac -= np.round(dfrac)
    return dfrac @ cell


# ---------- Grid utils (non-PBC only) ----------

def _grid_key(p: np.ndarray, inv_h: float, origin: np.ndarray) -> Tuple[int, int, int]:
    return tuple(((p - origin) * inv_h).astype(np.int64))  # type: ignore[return-value]


# ---------- Topology (built once per frame) ----------

def build_adjacency_nonpbc(symbols: Sequence[str], coords: np.ndarray, bond_detect_factor: float,
                           c_const: float = 0.3, bo_threshold: float = 0.2):
    N = len(symbols)
    radii = np.array([COVALENT_RADII.get(s, 0.77) for s in symbols], dtype=float)
    if N == 0:
        return [set()], {}, radii
    max_r = float(np.max(radii))
    h = bond_detect_factor * 2.0 * max_r + 1e-6
    inv_h = 1.0 / h
    origin = coords.min(axis=0) - 1e-6

    buckets: dict[Tuple[int, int, int], List[int]] = defaultdict(list)
    keys = np.empty((N, 3), dtype=np.int64)
    for i in range(N):
        k = _grid_key(coords[i], inv_h, origin)
        keys[i] = k
        buckets[k].append(i)

    adj: List[set[int]] = [set() for _ in range(N)]
    edge_len: dict[Tuple[int, int], float] = {}
    edge_order: dict[Tuple[int, int], int] = {}
    neighbor_offsets = [(dx, dy, dz) for dx in (-1, 0, 1) for dy in (-1, 0, 1) for dz in (-1, 0, 1)]
    for i in range(N):
        ri = radii[i]
        pi = coords[i]
        ki = tuple(keys[i])
        for dx, dy, dz in neighbor_offsets:
            k2 = (ki[0] + dx, ki[1] + dy, ki[2] + dz)
            if k2 not in buckets:
                continue
            for j in buckets[k2]:
                if j <= i:
                    continue
                rj = radii[j]
                pj = coords[j]
                rij2 = float(np.dot(pj - pi, pj - pi))
                rx = math.sqrt(rij2)
                r0 = (ri + rj)
                # Linus Pauling bond order
                bo = math.exp((r0 - rx) / float(c_const))
                if bo <= bo_threshold:
                    continue
                # classify bond order
                rounded = int(math.floor(bo + 0.5))
                order = 1 if rounded <= 1 else min(3, rounded)
                adj[i].add(j)
                adj[j].add(i)
                edge_len[(i, j)] = rx
                edge_order[(i, j)] = order
    return adj, edge_len, radii, edge_order


def build_adjacency_pbc(symbols: Sequence[str], coords: np.ndarray, cell: np.ndarray, bond_detect_factor: float,
                        c_const: float = 0.3, bo_threshold: float = 0.2):
    N = len(symbols)
    radii = np.array([COVALENT_RADII.get(s, 0.77) for s in symbols], dtype=float)
    if N == 0:
        return [set()], {}, radii
    inv_cell_T = np.linalg.inv(cell).T
    adj: List[set[int]] = [set() for _ in range(N)]
    edge_len: dict[Tuple[int, int], float] = {}
    edge_order: dict[Tuple[int, int], int] = {}
    for i in range(N):
        ri = radii[i]
        for j in range(i + 1, N):
            rj = radii[j]
            dvec = coords[j] - coords[i]
            dvec = mic_delta(dvec, cell, inv_cell_T)
            d2 = float(np.dot(dvec, dvec))
            rx = math.sqrt(d2)
            r0 = (ri + rj)
            bo = math.exp((r0 - rx) / float(c_const))
            if bo <= bo_threshold:
                continue
            rounded = int(math.floor(bo + 0.5))
            order = 1 if rounded <= 1 else min(3, rounded)
            adj[i].add(j)
            adj[j].add(i)
            edge_len[(i, j)] = rx
            edge_order[(i, j)] = order
    return adj, edge_len, radii, edge_order


# ---------- Fast torsion detection ----------

def get_bridges(adj: Sequence[set[int]]):
    sys.setrecursionlimit(max(100000, 10 * len(adj) + 100))
    n = len(adj)
    timer = 0
    tin = [-1] * n
    low = [-1] * n
    bridges: set[Tuple[int, int]] = set()

    def dfs(v: int, p: int):
        nonlocal timer
        tin[v] = low[v] = timer
        timer += 1
        for to in adj[v]:
            if to == p:
                continue
            if tin[to] != -1:
                low[v] = min(low[v], tin[to])
            else:
                dfs(to, v)
                low[v] = min(low[v], low[to])
                if low[to] > tin[v]:
                    a, b = (v, to) if v < to else (to, v)
                    bridges.add((a, b))

    for v in range(n):
        if tin[v] == -1:
            dfs(v, -1)
    return bridges


def _prefer_neighbor(a: int, exclude_b: int, adj: Sequence[set[int]], symbols: Sequence[str]) -> Optional[int]:
    heavy = [n for n in adj[a] if n != exclude_b and symbols[n] != 'H']
    if heavy:
        return heavy[0]
    for n in adj[a]:
        if n != exclude_b:
            return n
    return None


def get_rotatable_torsions_fast(adj: Sequence[set[int]], edge_len: dict[Tuple[int, int], float],
                                radii: np.ndarray, symbols: Sequence[str], mult_bond_factor: float,
                                edge_order: Optional[dict[Tuple[int, int], int]] = None):
    """
    Prefer bridge edges; if none, fall back to all internal bonds (deg>1) excluding suspected multiple bonds.
    Returns List[(n1,a,b,n2)]
    """
    N = len(adj)
    deg = [len(adj[i]) for i in range(N)]
    bridges = get_bridges(adj)

    def build_from_edges(edges: Iterable[Tuple[int, int]]):
        torsions: list[Tuple[int, int, int, int]] = []
        for (a, b) in edges:
            if a > b:
                a, b = b, a
            d = edge_len.get((a, b))
            if d is None:
                continue
            if deg[a] <= 1 or deg[b] <= 1:
                continue
            # Exclude multiple bonds: prefer bond order if available, otherwise fallback to distance heuristic
            if edge_order is not None:
                if edge_order.get((a, b), 1) >= 2:
                    continue
            else:
                if d < mult_bond_factor * (radii[a] + radii[b]):
                    continue
            n1 = _prefer_neighbor(a, b, adj, symbols)
            n2 = _prefer_neighbor(b, a, adj, symbols)
            if n1 is None or n2 is None:
                continue
            torsions.append((n1, a, b, n2))
        return torsions

    torsions = build_from_edges(bridges)
    if not torsions:
        torsions = build_from_edges(edge_len.keys())
    return torsions


def get_local_subtree_adj(adj: Sequence[set[int]], start_atom: int, exclude_atom: int,
                          max_subtree_atoms: Optional[int] = None) -> set[int]:
    visited: set[int] = set()
    queue: deque[int] = deque([start_atom])
    while queue:
        atom = queue.popleft()
        if atom in visited:
            continue
        visited.add(atom)
        for ni in adj[atom]:
            if ni == exclude_atom or ni in visited:
                continue
            if max_subtree_atoms is not None and len(visited) >= max_subtree_atoms:
                return visited
            queue.append(ni)
    return visited


# ---------- Guards ----------

def bonds_within_range_nonpbc(coords: np.ndarray, bond_pairs: Sequence[Tuple[int, int]], radii: np.ndarray,
                              min_factor: float, max_factor: Optional[float], detect_factor: float) -> bool:
    maxf = detect_factor if max_factor is None else float(max_factor)
    minf = float(min_factor) if (min_factor is not None) else 0.0
    for (i, j) in bond_pairs:
        d = float(np.linalg.norm(coords[i] - coords[j]))
        ref = radii[i] + radii[j]
        if d > maxf * ref:
            return False
        if minf > 0.0 and d < minf * ref:
            return False
    return True


def bonds_within_range_pbc(coords: np.ndarray, bond_pairs: Sequence[Tuple[int, int]], radii: np.ndarray,
                           min_factor: float, max_factor: Optional[float], detect_factor: float,
                           cell: np.ndarray, inv_cell_T: np.ndarray) -> bool:
    maxf = detect_factor if max_factor is None else float(max_factor)
    minf = float(min_factor) if (min_factor is not None) else 0.0
    for (i, j) in bond_pairs:
        dvec = mic_delta(coords[j] - coords[i], cell, inv_cell_T)
        d = float(np.linalg.norm(dvec))
        ref = radii[i] + radii[j]
        if d > maxf * ref:
            return False
        if minf > 0.0 and d < minf * ref:
            return False
    return True


def nonbond_clash_free_fast_nonpbc(coords: np.ndarray, radii: np.ndarray, bonded_set: set[Tuple[int, int]],
                                   min_factor: float) -> bool:
    N = coords.shape[0]
    if N == 0 or min_factor <= 0.0:
        return True
    max_r = float(np.max(radii))
    h = min_factor * 2.0 * max_r + 1e-6
    inv_h = 1.0 / h
    origin = coords.min(axis=0) - 1e-6

    buckets: dict[Tuple[int, int, int], List[int]] = defaultdict(list)
    keys = np.empty((N, 3), dtype=np.int64)
    for i in range(N):
        k = _grid_key(coords[i], inv_h, origin)
        keys[i] = k
        buckets[k].append(i)

    neighbor_offsets = [(dx, dy, dz) for dx in (-1, 0, 1) for dy in (-1, 0, 1) for dz in (-1, 0, 1)]
    for i in range(N):
        ri = radii[i]
        pi = coords[i]
        ki = tuple(keys[i])
        for dx, dy, dz in neighbor_offsets:
            k2 = (ki[0] + dx, ki[1] + dy, ki[2] + dz)
            if k2 not in buckets:
                continue
            for j in buckets[k2]:
                if j <= i:
                    continue
                if (i, j) in bonded_set:
                    continue
                rj = radii[j]
                pj = coords[j]
                rij2 = float(np.dot(pj - pi, pj - pi))
                cutoff = min_factor * (ri + rj)
                if rij2 < cutoff * cutoff:
                    return False
    return True


def nonbond_clash_free_fast_pbc(coords: np.ndarray, radii: np.ndarray, bonded_set: set[Tuple[int, int]],
                                min_factor: float, cell: np.ndarray, inv_cell_T: np.ndarray) -> bool:
    N = coords.shape[0]
    if N == 0 or min_factor <= 0.0:
        return True
    for i in range(N):
        ri = radii[i]
        for j in range(i + 1, N):
            if (i, j) in bonded_set:
                continue
            rj = radii[j]
            dvec = mic_delta(coords[j] - coords[i], cell, inv_cell_T)
            d2 = float(np.dot(dvec, dvec))
            cutoff = min_factor * (ri + rj)
            if d2 < cutoff * cutoff:
                return False
    return True


# ---------- Perturbation ----------

def perturb_coords_fast(coords: np.ndarray, adj: Sequence[set[int]], torsions: Sequence[Tuple[int, int, int, int]],
                        degrees_range: Tuple[float, float], num_torsions: int,
                        sigma: float, local_mode: bool, max_subtree_atoms: Optional[int],
                        pbc_active: bool, cell: Optional[np.ndarray], inv_cell_T: Optional[np.ndarray]) -> np.ndarray:
    if torsions:
        chosen = random.sample(list(torsions), min(len(torsions), num_torsions))
        for (n1, a, b, n2) in chosen:
            if local_mode:
                subtree_atoms = get_local_subtree_adj(adj, b, a, max_subtree_atoms)
            else:
                subtree_atoms = get_local_subtree_adj(adj, b, a, None)
            angle = random.uniform(*degrees_range)
            if subtree_atoms:
                if pbc_active and cell is not None and inv_cell_T is not None:
                    dvec = mic_delta(coords[b] - coords[a], cell, inv_cell_T)
                    axis_p1 = coords[a]
                    axis_p2 = coords[a] + dvec
                else:
                    axis_p1 = coords[a]
                    axis_p2 = coords[b]
                coords = rotate_coords(coords, subtree_atoms, axis_p1, axis_p2, angle)
    if sigma > 0:
        coords = coords + np.random.normal(0.0, sigma, size=coords.shape)
    return coords


# ---------- Public API ----------

def process_single(symbols: Sequence[str],
                   coords: np.ndarray,
                   cell: Optional[np.ndarray],
                   params: TorsionGuardParams) -> list[tuple]:

    # Decide PBC activation for this frame
    has_cell = cell is not None
    if params.pbc_mode == "yes":
        pbc_active = bool(has_cell)
        if not pbc_active:
            print("[WARN] pbc=yes but no Lattice found in frame; falling back to non-PBC.", file=sys.stderr)
    elif params.pbc_mode == "no":
        pbc_active = False
    else:  # "auto"
        pbc_active = bool(has_cell)

    inv_cell_T = np.linalg.inv(cell).T if (pbc_active and has_cell and cell is not None) else None

    # Build topology once (fixed for this frame). Multiple disconnected molecules are allowed.
    if pbc_active:
        assert cell is not None
        adj, edge_len, radii, edge_order = build_adjacency_pbc(
            symbols, coords, cell,
            bond_detect_factor=params.bond_detect_factor,
            c_const=float(params.bo_c_const), bo_threshold=float(params.bo_threshold)
        )
    else:
        adj, edge_len, radii, edge_order = build_adjacency_nonpbc(
            symbols, coords,
            bond_detect_factor=params.bond_detect_factor,
            c_const=float(params.bo_c_const), bo_threshold=float(params.bo_threshold)
        )

    bond_pairs = [(a, b) if a < b else (b, a) for (a, b) in edge_len.keys()]
    bonded_set = set(bond_pairs)

    # Torsion set (always fast method with fallback to internal bonds)
    torsions = get_rotatable_torsions_fast(adj, edge_len, radii, symbols, params.mult_bond_factor, edge_order=edge_order)
    local_mode_flag = len(symbols) > params.local_mode_cutoff_atoms

    results: list[tuple] = []
    for _ in range(int(params.perturb_per_frame)):
        new_coords = coords.copy()
        success = False
        for attempt in range(int(params.max_retries_per_frame) + 1):
            scale = 1.0 if attempt == 0 else (0.5 ** attempt)
            ang_lo, ang_hi = params.torsion_range_deg
            angle_range = (float(ang_lo) * scale, float(ang_hi) * scale)
            sigma_scaled = float(params.gaussian_sigma) * scale

            new_coords = coords.copy()
            new_coords = perturb_coords_fast(
                new_coords, adj, torsions,
                angle_range, int(params.max_torsions_per_conf),
                sigma_scaled, local_mode_flag, int(params.local_torsion_max_subtree),
                pbc_active, cell if cell is not None else None, inv_cell_T
            )

            # Guards: only original bonded pairs are enforced; non-bonded pairs kept apart
            if pbc_active and cell is not None and inv_cell_T is not None:
                if not bonds_within_range_pbc(
                    new_coords, bond_pairs, radii,
                    float(params.bond_keep_min_factor), params.bond_keep_max_factor, float(params.bond_detect_factor),
                    cell, inv_cell_T
                ):
                    continue
                if not nonbond_clash_free_fast_pbc(
                    new_coords, radii, bonded_set, float(params.nonbond_min_factor),
                    cell, inv_cell_T
                ):
                    continue
            else:
                if not bonds_within_range_nonpbc(
                    new_coords, bond_pairs, radii,
                    float(params.bond_keep_min_factor), params.bond_keep_max_factor, float(params.bond_detect_factor)
                ):
                    continue
                if not nonbond_clash_free_fast_nonpbc(
                    new_coords, radii, bonded_set, float(params.nonbond_min_factor)
                ):
                    continue

            success = True
            break

        if not success:
            new_coords = coords.copy()

        if pbc_active and cell is not None and inv_cell_T is not None:
            new_coords = wrap_to_cell(new_coords, cell, inv_cell_T)
        else:
            new_coords = center_in_box(new_coords, float(params.nonpbc_box_size))

        results.append((list(symbols), new_coords, cell, pbc_active))

    return results




from __future__ import annotations
import datetime as dt

from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship, column_property, backref
from sqlalchemy import String, Integer, DateTime, ForeignKey, Text, JSON, Boolean, Float, select, func, CTE, Index


class Base(DeclarativeBase):
    pass



class StorageRef(Base):
    __tablename__ = "storage_refs"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    scheme: Mapped[str] = mapped_column(String(10), nullable=False)  # file:// | cas://
    uri: Mapped[str] = mapped_column(Text, nullable=False)           # 路径或CAS相对路径
    content_hash: Mapped[str] = mapped_column(String(64), index=True)
    size: Mapped[int | None] = mapped_column(Integer)
    mtime: Mapped[dt.datetime | None] = mapped_column(DateTime)
    exist_flag: Mapped[bool] = mapped_column(Boolean, default=True)
    created_at: Mapped[dt.datetime] = mapped_column(DateTime, default=dt.datetime.utcnow)
class Tag(Base):
    __tablename__ = "tags"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    # SQLite 下用 NOCASE 保证 name 唯一且大小写不敏感
    name: Mapped[str] = mapped_column(
        String(64, collation="NOCASE"), nullable=False, unique=True, index=True
    )
    color: Mapped[str] = mapped_column(String(9), default="#D1E9FF")  # 可选：HEX 颜色
    notes: Mapped[str] = mapped_column(Text, default="")
    created_at: Mapped[dt.datetime] = mapped_column(DateTime, default=dt.datetime.utcnow)

    # 反向关系
    models: Mapped[set["ModelVersion"]] = relationship(
        "ModelVersion",
        secondary="model_version_tags",
        back_populates="tags",
        collection_class=set,      # 防止重复插入
        lazy="selectin",
    )
class ModelVersionTag(Base):
    __tablename__ = "model_version_tags"

    model_version_id: Mapped[int] = mapped_column(
        ForeignKey("model_versions.id", ondelete="CASCADE"),
        primary_key=True
    )
    tag_id: Mapped[int] = mapped_column(
        ForeignKey("tags.id", ondelete="CASCADE"),
        primary_key=True
    )
    created_at: Mapped[dt.datetime] = mapped_column(DateTime, default=dt.datetime.utcnow)

    # 额外索引：提高按 tag 查模型的性能
    __table_args__ = (
        Index("ix_mvt_tag_id", "tag_id"),
    )
class ModelVersion(Base):
    __tablename__ = "model_versions"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(String(80))

    # 归属项目 → 删项目时，模型级联删除
    project_id: Mapped[int] = mapped_column(
        ForeignKey("projects.id", ondelete="CASCADE"),
        index=True
    )

    model_type: Mapped[str] = mapped_column(String(20))  # NEP | DeepMD | ...
    data_size: Mapped[int] = mapped_column(Integer)
    energy: Mapped[float] = mapped_column(Float)
    force: Mapped[float] = mapped_column(Float)
    virial: Mapped[float] = mapped_column(Float)

    model_path: Mapped[str] = mapped_column(Text, nullable=False)
    notes: Mapped[str] = mapped_column(Text, default="")

    # 自引用（版本继承/分支）→ 删父版本时，子版本级联删除
    parent_id: Mapped[int | None] = mapped_column(
        ForeignKey("model_versions.id", ondelete="CASCADE"),
        index=True
    )

    created_at: Mapped[dt.datetime] = mapped_column(DateTime, default=dt.datetime.utcnow)

    # 关系：归属项目
    project: Mapped["Project"] = relationship(
        back_populates="model_versions",
        foreign_keys=[project_id],
        passive_deletes=True,      # 交给 DB ondelete 处理
    )

    # 关系：父/子版本树
    parent: Mapped["ModelVersion | None"] = relationship(
        remote_side=[id],
        backref=backref(
            "children",
            cascade="all, delete-orphan",
            passive_deletes=True
        ),
        single_parent=True,        # delete-orphan 需要
        passive_deletes=True
    )

    # 多对多标签（这里不动）
    tags: Mapped[set["Tag"]] = relationship(
        "Tag",
        secondary="model_version_tags",
        back_populates="models",
        collection_class=set,
        lazy="selectin",
        passive_deletes=True
    )


class Project(Base):
    __tablename__ = "projects"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(String(200), unique=True, nullable=False)
    notes: Mapped[str] = mapped_column(Text, default="")

    # 当前激活模型 → 被删时置空，避免 FK 失败
    # active_model_version_id: Mapped[int | None] = mapped_column(
    #     ForeignKey("model_versions.id", ondelete="SET NULL"),
    #     nullable=True
    # )

    created_at: Mapped[dt.datetime] = mapped_column(DateTime, default=dt.datetime.utcnow)
    updated_at: Mapped[dt.datetime] = mapped_column(
        DateTime, default=dt.datetime.utcnow, onupdate=dt.datetime.utcnow
    )

    # 自引用项目树 → 删父项目时子项目级联删除
    parent_id: Mapped[int | None] = mapped_column(
        ForeignKey("projects.id", ondelete="CASCADE"),
        index=True
    )

    # 项目树关系
    parent: Mapped["Project | None"] = relationship(
        remote_side=[id],
        backref=backref(
            "children",
            cascade="all, delete-orphan",
            passive_deletes=True
        ),
        single_parent=True,
        passive_deletes=True
    )

    # 项目下的所有模型 → 删项目时“模型”也删（双保险：ORM + DB）
    model_versions: Mapped[list["ModelVersion"]] = relationship(
        back_populates="project",
        foreign_keys="ModelVersion.project_id",
        cascade="all, delete-orphan",
        passive_deletes=True
    )

    # 如需：激活版本关系（可保持注释，或启用）
    # active_model_version: Mapped["ModelVersion | None"] = relationship(
    #     "ModelVersion",
    #     foreign_keys=[active_model_version_id],
    #     uselist=False,
    #     post_update=True,
    #     passive_deletes=True
    # )

    model_num: Mapped[int] = column_property(
        select(func.count(ModelVersion.id))
        .where(ModelVersion.project_id == id)
        .scalar_subquery()
    )


class Event(Base):
    __tablename__ = "events"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    entity_type: Mapped[str] = mapped_column(String(50))
    entity_id: Mapped[int] = mapped_column(Integer)
    action: Mapped[str] = mapped_column(String(50))
    payload_json: Mapped[dict] = mapped_column(JSON, default=dict)
    created_at: Mapped[dt.datetime] = mapped_column(DateTime, default=dt.datetime.utcnow)
    actor: Mapped[str] = mapped_column(String(80), default="system")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/10/18 15:31
# @Author  : 兵
# @email    : 1747193328@qq.com
import os
import threading
import re
import traceback
from functools import cached_property
from pathlib import Path
import numpy as np
from PySide6.QtCore import QObject, Signal
from loguru import logger
from typing import Any, Callable, Optional
from numpy import bool_
import numpy.typing as npt
from NepTrainKit import utils
from NepTrainKit.config import Config
from NepTrainKit.core import Structure, MessageManager
from NepTrainKit.core.calculator import NEPProcess, run_nep_calculator, NepCalculator
from NepTrainKit.core.io.utils import read_nep_out_file, parse_array_by_atomnum,get_rmse
from NepTrainKit.core.types import Brushes, SearchType, NepBackend


def pca(X, n_components=None):
    """
    执行主成分分析 (PCA)，只返回降维后的数据
    """
    n_samples, n_features = X.shape

    # 1. 计算均值并中心化数据
    mean = np.mean(X, axis=0)
    X_centered = X - mean
    #樊老师说不用处理 就不减去均值了
    # 但是我还不确定哪种好 还是保持现状把
    # X_centered = X


    # 3. 计算协方差矩阵
    cov_matrix = np.dot(X_centered.T, X_centered) / (n_samples - 1)

    # 4. 计算特征值和特征向量
    eigenvalues, eigenvectors = np.linalg.eigh(cov_matrix)

    # 5. 特征值和特征向量按降序排列
    idx = np.argsort(eigenvalues)[::-1]
    eigenvalues = eigenvalues[idx]
    eigenvectors = eigenvectors[:, idx]

    # 6. 确定要保留的主成分数量
    if n_components is None:
        n_components = n_features
    elif n_components > n_features:
        n_components = n_features

    # 7. 将数据投影到前n_components个主成分上 (降维)
    X_pca = np.dot(X_centered, eigenvectors[:, :n_components])

    return X_pca.astype(np.float32)



class DataBase:
    """
    优化后的 DataBase 类，对列表进行封装，支持根据索引删除结构和回退。
    使用布尔掩码管理活动/删除状态，减少列表操作开销。
    """
    def __init__(self, data_list:list[Any]|npt.NDArray[Any]):
        """Initialize with a NumPy array."""
        self._data = np.asarray(data_list)
        # 布尔掩码：True 表示活跃，False 表示已删除
        self._active_mask = np.ones(len(self._data), dtype=bool)
        # 历史记录栈，存储每次删除的掩码变化
        self._history = []
    @property
    def mask_array(self):
        return self._active_mask

    @property
    def num(self) -> int:
        """返回当前活跃数据的数量"""
        return np.sum(self._active_mask)
    @property
    def all_data(self) -> npt.NDArray[Any]:
        return self._data
    @property
    def now_data(self) -> npt.NDArray[Any]:
        """返回当前活跃数据"""
        return self._data[self._active_mask]

    @property
    def remove_data(self) -> npt.NDArray[Any]:
        """返回所有已删除的数据"""
        return self._data[~self._active_mask]

    @property
    def now_indices(self) -> npt.NDArray[np.int32]:
        """返回当前活跃数据的索引下标"""
        return np.where(self._active_mask)[0]

    @property
    def remove_indices(self) -> npt.NDArray[np.int32]:
        """返回已删除数据的索引下标"""
        return np.where(~self._active_mask)[0]

    def remove(self, indices)->None:
        idx = np.unique(np.asarray(indices, dtype=int) if not isinstance(indices, int) else [indices])
        idx = idx[(idx >= 0) & (idx < len(self._data))]
        if len(idx) == 0:
            return
        self._history.append(idx)  # 存储删除的索引
        self._active_mask[idx] = False

    def revoke(self)->None:
        if self._history:
            last_indices = self._history.pop()
            self._active_mask[last_indices] = True

    def __getitem__(self, item):
        """直接索引活跃数据集"""
        return self.now_data[item]


class NepData:
    """
    structure_data 结构性质数据点
    group_array 结构的组号 标记数据点对应结构在train.xyz中的下标
    title 能量 力 等 用于画图axes的标题

    """
    title:str
    def __init__(self,
                 data_list:list[Any]|npt.NDArray[Any],
                 group_list:int|list[int]=1,
                 **kwargs ):
        if isinstance(data_list,(list )):
            data_list=np.array(data_list)

        self.data = DataBase(data_list)
        if isinstance(group_list,int):
            group = np.arange(data_list.shape[0],dtype=np.uint32)
            self.group_array=DataBase(group)
        else:
            group = np.arange(len(group_list),dtype=np.uint32 )
            self.group_array=DataBase(group.repeat(group_list))

        for key,value in kwargs.items():
            setattr(self,key,value)
    @property
    def num(self)->int:
        return self.data.num
    @cached_property
    def cols(self)->int:
        """
        将列数除以2 前面是nep 后面是dft
        """
        if self.now_data.shape[0]==0:
            #数据为0
            return 0
        index = self.now_data.shape[1] // 2
        return index
    @property
    def now_data(self) -> npt.NDArray[Any]:
        """
        返回当前数据
        """
        return self.data.now_data
    @property
    def now_indices(self) -> npt.NDArray[np.int32]:
        return self.data.now_indices
    @property
    def all_data(self) -> npt.NDArray[Any]:
        return self.data.all_data
    def is_visible(self,index) -> bool:
        if self.data.all_data.size == 0:
            return False
        return self.data.mask_array[index].all()
    @property
    def remove_data(self) -> npt.NDArray[Any]:
        """返回删除的数据"""

        return self.data.remove_data

    def convert_index(self,index_list:list[int]|int) -> npt.NDArray[np.int32]:
        """
        传入结构的原始下标 然后转换成现在已有的
        这个主要是映射force等性质的下标 或者其他一对多的性质
        对于一对一的比如，转换后会将index_list按照原始下标排序
        输入[np.int64(98), np.int64(9), np.int64(42), np.int64(141), np.int64(79), np.int64(56)]
        输出[  9  42  56  79  98 141]
        """
        if isinstance(index_list,(int,np.number)):
            index_list=[index_list]
        return np.where(np.isin(self.group_array.all_data,index_list))[0]



    def remove(self,remove_index:list[int]|int):
        """
        根据index删除
        remove_index 结构的原始下标
        """
        remove_indices=self.convert_index(remove_index)

        self.data.remove(remove_indices)
        self.group_array.remove(remove_indices)

    def revoke(self):
        """将上一次删除的数据恢复"""
        self.data.revoke()
        self.group_array.revoke()

    def get_rmse(self)->float:
        if not self.cols:
            return 0
        return get_rmse(self.now_data[:, 0:self.cols],self.now_data[:, self.cols: ])
        # return np.sqrt(((self.now_data[:, 0:self.cols] - self.now_data[:, self.cols: ]) ** 2).mean( ))

    def get_formart_rmse(self)->str:
        rmse=self.get_rmse()
        if self.title =="energy":
            unit="meV/atom"
            rmse*=1000
        elif self.title =="force":
            unit="meV/Å"
            rmse*=1000
        elif self.title =="virial":
            unit="meV/atom"
            rmse*=1000
        elif self.title =="stress":
            unit="MPa"
            rmse*=1000
        elif "Polar" in self.title:
            unit="(m.a.u./atom)"
            rmse*=1000
        elif "dipole" == self.title:
            unit="(m.a.u./atom)"
            rmse*=1000
        elif "spin" ==self.title:
            unit = "meV/μB"
            rmse*=1000
        else:
            return ""
        return f"{rmse:.2f} {unit}"

    def get_max_error_index(self,nmax)->list[int]:
        """
        返回nmax个最大误差的下标
        这个下标是结构的原始下标
        """
        error = np.sum(np.abs(self.now_data[:, 0:self.cols] - self.now_data[:, self.cols: ]), axis=1)
        rmse_max_ids = np.argsort(-error)
        structure_index =self.group_array.now_data[rmse_max_ids]
        index,indices=np.unique(structure_index,return_index=True)

        return structure_index[np.sort(indices)][:nmax].tolist()




class NepPlotData(NepData):

    def __init__(self,data_list,**kwargs ):
        super().__init__(data_list,**kwargs )
        self.x_cols=slice(self.cols,None)
        self.y_cols=slice(None,self.cols )

    @property
    def x(self) -> npt.NDArray[Any]:
        """
        这里返回的是展平之后的数据 方便直接画图
        """
        if self.cols==0:
            return self.now_data
        return self.now_data[ : ,self.x_cols].ravel()
    @property
    def y(self) -> npt.NDArray[Any]:
        if self.cols==0:
            return self.now_data
        return self.now_data[ : , self.y_cols].ravel()


    @property
    def structure_index(self):
        """
        这里根据列数 将group做一个扩充 传入到画图的data里
        为了将散点将结构下标映射起来
        """
        return self.group_array[ : ].repeat(self.cols)
class DPPlotData(NepData):


    def __init__(self,data_list,**kwargs ):
        super().__init__(data_list,**kwargs )
        self.x_cols=slice(None,self.cols)
        self.y_cols=slice(self.cols,None)

    @property
    def x(self):
        if self.cols==0:
            return self.now_data
        return self.now_data[ : ,self.x_cols].ravel()
    @property
    def y(self):
        if self.cols==0:
            return self.now_data
        return self.now_data[ : , self.y_cols].ravel()



    # def all_x(self):
    #     if self.cols==0:
    #         return self.all_data
    #     return self.all_data[ : ,:self.cols].ravel()
    # @property
    # def all_y(self):
    #     if self.cols==0:
    #         return self.all_data
    #     return self.all_data[ : , self.cols:].ravel()
    @property
    def structure_index(self):
        return self.group_array[ : ].repeat(self.cols)

class StructureData(NepData):

    @utils.timeit
    def get_all_config(self,search_type:SearchType=None)->list[str]:
        """
        获取所有结构的某个属性 用于搜索
        :param search_type:SearchType
        :return:每个结构的属性值
        """
        if search_type is None:
            search_type=SearchType.TAG
        if search_type==SearchType.TAG:
            return [structure.tag for structure in self.now_data]
        elif search_type==SearchType.FORMULA:
            return [structure.formula for structure in self.now_data]
        else:
            MessageManager.send_warning_message(f"no such search_type:{search_type}")
            return []
    def search_config(self,config:str,search_type:SearchType):
        """
        根据传入的config 对结构的属性值进行匹配
        这里使用了正则 可以使用正则支持的语法
        :param config:
        :param search_type:
        :return:符合搜索的结构的下标
        """
        if search_type==SearchType.TAG:

            result_index=[i for i ,structure in enumerate(self.now_data) if re.search(config, structure.tag)]
        elif search_type==SearchType.FORMULA:
            result_index=[i for i ,structure in enumerate(self.now_data) if re.search(config, structure.formula)]
        else:
            MessageManager.send_warning_message(f"no such search_type:{search_type}")
            return []
        return self.group_array[result_index].tolist()


class ResultData(QObject):
    #通知界面更新训练集的数量情况
    updateInfoSignal = Signal( )
    #加载训练集结束后发出的信号
    loadFinishedSignal = Signal()
    atoms_num_list: npt.NDArray
    _atoms_dataset: StructureData

    def __init__(self,
                 nep_txt_path:Path|str,
                 data_xyz_path:Path|str,
                 descriptor_path:Path|str,
                 calculator_factory: Optional[Callable[[str], Any]] = None,
                 import_options: Optional[dict] = None):
        super().__init__()
        self.load_flag=False
        # cooperative cancel for long-running loads
        self.cancel_event = threading.Event()

        self.descriptor_path=Path(descriptor_path)
        self.data_xyz_path=Path(data_xyz_path)
        self.nep_txt_path=Path(nep_txt_path)
        #存储选中结构的真实下标
        self.select_index=set()
        # Optional pre-fetched structures to skip IO in load_structures
        self._prefetched_structures: Optional[list[Structure]] = None
        # Optional importer options forwarded to importers.import_structures
        self._import_options: dict = dict(import_options or {})
        self.calculator_factory=calculator_factory

    def request_cancel(self):
        """Request cooperative cancel during load. Also forward to calculator."""
        self.cancel_event.set()
        try:
            if hasattr(self, "nep_calc") and self.nep_calc is not None:
                self.nep_calc.cancel()
        except Exception:
            pass

    def reset_cancel(self):
        self.cancel_event.clear()
    @utils.timeit
    def load_structures(self):
        """
        加载结构xyz文件
        :return:
        """
        # If structures were provided upfront, use them; otherwise parse from file
        if self._prefetched_structures is not None:
            structures = self._prefetched_structures
        else:
            # Unified path: delegate to importers for all formats, including EXTXYZ.
            # ExtxyzImporter internally uses Structure.iter_read_multiple with cancel support.
            from NepTrainKit.core.io import importers as _imps
            opts = dict(self._import_options)
            opts.setdefault("cancel_event", self.cancel_event)
            structures = _imps.import_structures(self.data_xyz_path.as_posix(), **opts)
        self._atoms_dataset = StructureData(structures)
        self.atoms_num_list = np.array([len(struct) for struct in self.structure.now_data])

    def set_structures(self, structures: list[Structure]):
        """
        Provide pre-parsed structures so load_structures can skip file IO.
        """
        self._prefetched_structures = list(structures)

    def write_prediction(self):
        """
        程序将严格检测batch和结构数的关系，如果不是fullbatch，
        就会触发计算，通过写入一个nep.in文件 可以避免这种情况。
        在nep_cpu计算后，回调用该函数
        :return:
        """
        if self.atoms_num_list.shape[0] > 1000:
            #
            if not self.data_xyz_path.with_name("nep.in").exists():
                with open(self.data_xyz_path.with_name("nep.in"),
                          "w", encoding="utf8") as f:
                    f.write("prediction 1 ")

    def load(self ):
        """
        加载数据集的主函数
        :return:
        """
        try:
            # Calculator injection (default to NEP). Subclasses can pass in a factory for other ML potentials.
            if self.calculator_factory is None:
                self.nep_calc = NepCalculator(
                    model_file=self.nep_txt_path.as_posix(),
                    backend=NepBackend(Config.get("nep", "backend", "auto")),
                    batch_size=Config.getint("nep", "gpu_batch_size", 1000)
                )
            else:
                # Factory is responsible for creating a calculator compatible with this ResultData subclass
                try:
                    self.nep_calc = self.calculator_factory(self.nep_txt_path.as_posix())
                except Exception:
                    logger.debug(traceback.format_exc())
                    MessageManager.send_warning_message("Failed to create custom calculator; falling back to NEP.")
                    self.nep_calc = NepCalculator(
                        model_file=self.nep_txt_path.as_posix(),
                        backend=NepBackend(Config.get("nep", "backend", "auto")),
                        batch_size=Config.getint("nep", "gpu_batch_size", 1000)
                    )


            # If subclass overrides load_structures, defer to it; otherwise do cancel-aware read

            self.load_structures()
            if self._atoms_dataset.num!=0:



                if not self.cancel_event.is_set():
                    self._load_descriptors()
                if not self.cancel_event.is_set():
                    self._load_dataset()
                if not self.cancel_event.is_set():
                    self.load_flag=True
            else:
                MessageManager.send_warning_message("No structures were loaded.")

        except:
            logger.error(traceback.format_exc())

            MessageManager.send_error_message("load dataset error!")

        self.loadFinishedSignal.emit()
    def _load_dataset(self):
        """
        不同类型的势函数通过重写该函数实现不同的加载逻辑
        主要是加载结构性质
        :return:
        """
        raise NotImplementedError()

    @property
    def datasets(self) -> list["NepPlotData"]:
        raise NotImplementedError()

    @property
    def descriptor(self):
        return self._descriptor_dataset

    @property
    def num(self):
        return self._atoms_dataset.num
    @property
    def structure(self):
        return self._atoms_dataset

    def is_select(self,i):
        return i in self.select_index

    def select(self,indices:list[int]|int):
        """
        传入一个索引列表，将索引对应的结构标记为选中状态
        这个下标是结构在train.xyz中的索引
        :param indices: 索引
        :return:
        """


        # 统一转换为 NumPy 数组
        idx = np.asarray(indices, dtype=int) if not isinstance(indices, int) else np.array([indices])
        # 去重并过滤有效索引（在数据范围内且为活跃数据）
        idx = np.unique(idx)
        idx = idx[(idx >= 0) & (idx < len(self.structure.all_data)) & (self.structure.data.mask_array[idx])]
        # 批量添加到选中集合
        self.select_index.update(idx)

        self.updateInfoSignal.emit()

    def uncheck(self,_list:list[int]|int):
        """
        check_list 传入一个索引列表，将索引对应的结构标记为未选中状态
        这个下标是结构在train.xyz中的索引
        :param _list:
        :return:
        """
        if isinstance(_list,(int,np.number)):
            _list=[_list]
        for i in _list:
            if i in self.select_index:
                self.select_index.remove(i)

        self.updateInfoSignal.emit()

    def inverse_select(self):
        """
        根据现在选择的结构对先有训练集进行一个反选的操作
        :return:
        """
        active_indices = set(self.structure.data.now_indices.tolist())
        selected_indices = set(self.select_index)
        unselect = list(selected_indices)
        select = list(active_indices - selected_indices)
        if unselect:
            self.uncheck(unselect)
        if select:
            self.select(select)
    def get_selected_structures(self)->list[Structure]:
        """
        获取选中结构的list
        :return: list[Structure]
        """
        index=list(self.select_index)
        index = self.structure.convert_index(index)

        return self.structure.all_data[index].tolist()

    def export_selected_xyz(self,save_file_path):
        """
        导出当前选中的结构
        :param save_file_path: 保存文件路径，具体到文件 例如~/select.xyz
        :return:
        """
        index=list(self.select_index)
        try:
            with open(save_file_path,"w",encoding="utf8") as f:

                index=self.structure.convert_index(index)
                for structure in self.structure.all_data[index]:
                    structure.write(f)
            MessageManager.send_info_message(f"File exported to: {save_file_path}")
        except:
            MessageManager.send_info_message(f"An unknown error occurred while saving. The error message has been output to the log!")
            logger.error(traceback.format_exc())

    def export_model_xyz(self,save_path):
        """
        导出当前结构
        :param save_path: 保存路径
        被删除的导出到export_remove_model.xyz
        被保留的导出到export_good_model.xyz
        """
        try:

            with open(Path(save_path).joinpath("export_good_model.xyz"),"w",encoding="utf8") as f:
                for structure in self.structure.now_data:
                    structure.write(f)

            with open(Path(save_path).joinpath("export_remove_model.xyz"),"w",encoding="utf8") as f:
                for structure in self.structure.remove_data:
                    structure.write(f)


            MessageManager.send_info_message(f"File exported to: {save_path}")
        except:
            MessageManager.send_info_message(f"An unknown error occurred while saving. The error message has been output to the log!")
            logger.error(traceback.format_exc())


    def get_atoms(self,index ):
        """根据原始索引获取原子结构对象"""
        index=self.structure.convert_index(index)
        return self.structure.all_data[index][0]



    def remove(self,i):

        """
        在所有的dataset中删除某个索引对应的结构
        i是原始下标
        """
        self.structure.remove(i)
        for dataset in self.datasets:
            dataset.remove(i)
        self.updateInfoSignal.emit()

    @property
    def is_revoke(self):
        """
        判断是否有被删除的结构
        """
        return self.structure.remove_data.size!=0
    def revoke(self):
        """
        撤销到上一次的删除
        """
        self.structure.revoke()
        for dataset in self.datasets:
            dataset.revoke( )
        self.updateInfoSignal.emit()

    @utils.timeit
    def delete_selected(self ):
        """
        删除所有selected的结构
        """
        self.remove(list(self.select_index))
        self.select_index.clear()
        self.updateInfoSignal.emit()


    def _load_descriptors(self):
        """
        加载训练集的描述符
        :return:
        """

        if os.path.exists(self.descriptor_path):
            desc_array = read_nep_out_file(self.descriptor_path,dtype=np.float32,ndmin=2)

        else:
            desc_array = np.array([])

        if desc_array.size == 0:

            desc_array = self.nep_calc.get_structures_descriptor(self.structure.now_data.tolist())

            # desc_array=self.nep_calc_thread.func_result
            # desc_array = run_nep3_calculator_process(
            #     )

            if desc_array.size != 0:
                np.savetxt(self.descriptor_path, desc_array, fmt='%.6g')
        else:
            if desc_array.shape[0] == np.sum(self.atoms_num_list):
                # 原子描述符 需要计算结构描述符


                desc_array = parse_array_by_atomnum(desc_array, self.atoms_num_list, map_func=np.mean, axis=0)
            elif desc_array.shape[0] == self.atoms_num_list.shape[0]:
                # 结构描述符
                pass

            else:
                self.descriptor_path.unlink(True)
                return self._load_descriptors()

        if desc_array.size != 0:
            if desc_array.shape[1] > 2:
                try:
                    desc_array = pca(desc_array, 2)
                except:
                    MessageManager.send_error_message("PCA dimensionality reduction fails")
                    desc_array = np.array([])
        self._descriptor_dataset = NepPlotData(desc_array, title="descriptor")



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/28 22:45
# @Author  : 兵
# @email    : 1747193328@qq.com
import sys
from pathlib import Path
from typing import Any, Dict

from PySide6.QtGui import QIcon, QDoubleValidator, QIntValidator, QColor
from PySide6.QtWidgets import (QVBoxLayout, QFrame, QGridLayout,
                               QPushButton, QLineEdit, QWidget, QHBoxLayout, QFormLayout, QSizePolicy, QComboBox, QColorDialog)
from PySide6.QtCore import Signal, Qt, QUrl, QEvent
from qfluentwidgets import (
    MessageBoxBase,
    SpinBox,
    CaptionLabel,
    DoubleSpinBox,
    CheckBox,
    ProgressBar,
    ComboBox,
    FluentStyleSheet,
    FluentTitleBar, CompactDoubleSpinBox, TransparentToolButton, Dialog, ColorDialog,
    TitleLabel, HyperlinkLabel, RadioButton, LineEdit, FlowLayout, EditableComboBox, PrimaryDropDownPushButton,
    PrimaryPushButton, Flyout, InfoBarIcon, MessageBox, TextEdit, FluentIcon, PushButton, ToolTipFilter, ToolTipPosition
)
from qframelesswindow import FramelessDialog
import json
import os
from .button import TagPushButton, TagGroup

from NepTrainKit.core import MessageManager

from NepTrainKit import module_path

from NepTrainKit.utils import LoadingThread,call_path_dialog
from ..core.io.utils import get_xyz_nframe,  read_nep_out_file, get_rmse


class GetIntMessageBox(MessageBoxBase):
    """ Custom message box """

    def __init__(self, parent=None,tip=""):
        super().__init__(parent)
        self.titleLabel = CaptionLabel(tip, self)
        self.titleLabel.setWordWrap(True)
        self.intSpinBox = SpinBox(self)

        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self.intSpinBox)

        self.widget.setMinimumWidth(100 )
        self.intSpinBox.setMaximum(100000000)

class GetStrMessageBox(MessageBoxBase):
    """ Custom message box """

    def __init__(self, parent=None,tip=""):
        super().__init__(parent)
        self.titleLabel = CaptionLabel(tip, self)
        self.titleLabel.setWordWrap(True)
        self.lineEdit = LineEdit(self)

        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self.lineEdit)

        self.widget.setMinimumWidth(100 )


class SparseMessageBox(MessageBoxBase):
    """用于最远点取样的弹窗 """

    def __init__(self, parent=None,tip=""):
        super().__init__(parent)
        self.titleLabel = CaptionLabel(tip, self)
        self.titleLabel.setWordWrap(True)
        self._frame = QFrame(self)
        self.frame_layout=QGridLayout(self._frame)
        self.frame_layout.setContentsMargins(0,0,0,0)
        self.frame_layout.setSpacing(0)
        self.intSpinBox = SpinBox(self)

        self.intSpinBox.setMaximum(9999999)
        self.intSpinBox.setMinimum(0)
        self.doubleSpinBox = DoubleSpinBox(self)
        self.doubleSpinBox.setDecimals(5)
        self.doubleSpinBox.setMinimum(0)
        self.doubleSpinBox.setMaximum(10)

        self.frame_layout.addWidget(CaptionLabel("Max num", self),0,0,1,1)

        self.frame_layout.addWidget(self.intSpinBox,0,1,1,2)
        self.frame_layout.addWidget(CaptionLabel("Min distance", self),1,0,1,1)

        self.frame_layout.addWidget(self.doubleSpinBox,1,1,1,2)

        # region option: use current selection as FPS region
        self.regionCheck = CheckBox("Use current selection as region", self)
        self.regionCheck.setToolTip("When FPS sampling is performed in the designated area, the program will automatically deselect it, just click to delete!")
        self.regionCheck.installEventFilter(ToolTipFilter(self.regionCheck, 300, ToolTipPosition.TOP))

        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self._frame )
        self.viewLayout.addWidget(self.regionCheck)

        self.yesButton.setText('Ok')
        self.cancelButton.setText('Cancel')

        self.widget.setMinimumWidth(200)


class IndexSelectMessageBox(MessageBoxBase):
    """Dialog for selecting structures by index."""

    def __init__(self, parent=None, tip="Specify index or slice"):
        super().__init__(parent)
        self.titleLabel = CaptionLabel(tip, self)
        self.titleLabel.setWordWrap(True)
        self.indexEdit = LineEdit(self)
        self.checkBox = CheckBox("Use original indices", self)
        self.checkBox.setChecked(True)

        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self.indexEdit)
        self.viewLayout.addWidget(self.checkBox)

        self.yesButton.setText('Ok')
        self.cancelButton.setText('Cancel')
        self.widget.setMinimumWidth(200)


class RangeSelectMessageBox(MessageBoxBase):
    """Dialog for selecting structures by axis range."""

    def __init__(self, parent=None, tip="Specify x/y range"):
        super().__init__(parent)
        self.titleLabel = CaptionLabel(tip, self)
        self.titleLabel.setWordWrap(True)

        self._frame = QFrame(self)
        self.frame_layout = QGridLayout(self._frame)
        self.frame_layout.setContentsMargins(0, 0, 0, 0)
        self.frame_layout.setSpacing(2)

        self.xMinSpin = DoubleSpinBox(self)
        self.xMinSpin.setDecimals(6)
        self.xMinSpin.setRange(-1e8, 1e8)
        self.xMaxSpin = DoubleSpinBox(self)
        self.xMaxSpin.setDecimals(6)
        self.xMaxSpin.setRange(-1e8, 1e8)
        self.yMinSpin = DoubleSpinBox(self)
        self.yMinSpin.setDecimals(6)
        self.yMinSpin.setRange(-1e8, 1e8)
        self.yMaxSpin = DoubleSpinBox(self)
        self.yMaxSpin.setDecimals(6)
        self.yMaxSpin.setRange(-1e8, 1e8)

        self.logicCombo = ComboBox(self)
        self.logicCombo.addItems(["AND", "OR"])

        self.frame_layout.addWidget(CaptionLabel("X min", self), 0, 0)
        self.frame_layout.addWidget(self.xMinSpin, 0, 1)
        self.frame_layout.addWidget(CaptionLabel("X max", self), 0, 2)
        self.frame_layout.addWidget(self.xMaxSpin, 0, 3)
        self.frame_layout.addWidget(CaptionLabel("Y min", self), 1, 0)
        self.frame_layout.addWidget(self.yMinSpin, 1, 1)
        self.frame_layout.addWidget(CaptionLabel("Y max", self), 1, 2)
        self.frame_layout.addWidget(self.yMaxSpin, 1, 3)
        self.frame_layout.addWidget(CaptionLabel("Logic", self), 2, 0)
        self.frame_layout.addWidget(self.logicCombo, 2, 1, 1, 3)

        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self._frame)

        self.yesButton.setText('Ok')
        self.cancelButton.setText('Cancel')
        self.widget.setMinimumWidth(300)


class ArrowMessageBox(MessageBoxBase):
    """Dialog for selecting arrow display options."""

    def __init__(self, parent=None, props=None):
        super().__init__(parent)
        self.titleLabel = CaptionLabel("Vector property", self)
        self.titleLabel.setWordWrap(True)

        self._frame = QFrame(self)
        self.frame_layout = QGridLayout(self._frame)
        self.frame_layout.setContentsMargins(0, 0, 0, 0)
        self.frame_layout.setSpacing(2)

        self.propCombo = ComboBox(self)
        if props:
            self.propCombo.addItems(props)

        self.scaleSpin = DoubleSpinBox(self)
        self.scaleSpin.setDecimals(3)
        self.scaleSpin.setRange(0, 1000)
        self.scaleSpin.setValue(1.0)

        self.colorCombo = ComboBox(self)
        self.colorCombo.addItems(["viridis", "magma", "plasma", "inferno", "jet"])

        self.showCheck = CheckBox("Show arrows", self)
        self.showCheck.setChecked(True)

        self.frame_layout.addWidget(CaptionLabel("Property", self), 0, 0)
        self.frame_layout.addWidget(self.propCombo, 0, 1)
        self.frame_layout.addWidget(CaptionLabel("Scale", self), 1, 0)
        self.frame_layout.addWidget(self.scaleSpin, 1, 1)
        self.frame_layout.addWidget(CaptionLabel("Colormap", self), 2, 0)
        self.frame_layout.addWidget(self.colorCombo, 2, 1)
        self.frame_layout.addWidget(self.showCheck, 3, 0, 1, 2)

        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self._frame)

        self.yesButton.setText('Ok')
        self.cancelButton.setText('Cancel')
        self.widget.setMinimumWidth(250)
class InputInfoMessageBox(MessageBoxBase):


    def __init__(self, parent=None):
        super().__init__(parent)
        self.titleLabel = CaptionLabel("new structure info", self)
        self.titleLabel.setWordWrap(True)

        self._frame = QFrame(self)
        self.frame_layout = QGridLayout(self._frame)
        self.frame_layout.setContentsMargins(0, 0, 0, 0)
        self.frame_layout.setSpacing(2)

        self.keyEdit = LineEdit(self)
        self.valueEdit = LineEdit(self)
        self.frame_layout.addWidget(CaptionLabel("Key", self), 1, 0)
        self.frame_layout.addWidget(self.keyEdit, 1, 1, 1, 3)
        self.frame_layout.addWidget(CaptionLabel("Value", self), 2, 0)
        self.frame_layout.addWidget(self.valueEdit, 2, 1, 1, 3)
        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self._frame)

        self.yesButton.setText('Ok')
        self.cancelButton.setText('Cancel')
        self.widget.setMinimumWidth(100)
    def validate(self):
        if self.keyEdit.text().strip() != "":
            return True
        Flyout.create(
            icon=InfoBarIcon.INFORMATION,
            title='Tip',
            content="A valid value must be entered",
            target=self.keyEdit,
            parent=self,
            isClosable=True
        )
        return False
class EditInfoMessageBox(MessageBoxBase):
    """Dialog for editing structure information."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.titleLabel = CaptionLabel("Edit info", self)
        self.titleLabel.setWordWrap(True)
        self.new_tag_button = PrimaryPushButton(QIcon(":/images/src/images/copy_figure.svg"),
                                                         "Add new tag", self)
        self.new_tag_button.setMaximumWidth(200)
        self.new_tag_button.setObjectName("new_tag_button")
        self.new_tag_button.clicked.connect(self.new_tag)
        self.tag_group = TagGroup(parent=self)
        self.tag_group.tagRemovedSignal.connect(self.tag_removed)
        self.viewLayout.addWidget(self.new_tag_button)

        self.viewLayout.addWidget(self.tag_group)
        self.yesButton.setText('Ok')
        self.cancelButton.setText('Cancel')
        self.widget.setMinimumWidth(600)
        self.remove_tag=set()
        self.new_tag_info={}
    def new_tag(self):
        box = InputInfoMessageBox(self)
        if not box.exec():
            return
        key=box.keyEdit.text()
        value=box.valueEdit.text()

        if key.strip():
            self.add_tag(key.strip(),value)
    def init_tags(self, tags):
        for tag in tags:
            self.tag_group.add_tag(tag)
    def tag_removed(self,tag):
        if tag in self.new_tag_info.keys():
            self.new_tag_info.pop(tag)
        self.remove_tag.add(tag)
    def add_tag(self,tag,value):
        if self.tag_group.has_tag(tag):
            MessageManager.send_message_box(f"{tag} already exists, please delete it first")
            return
        self.new_tag_info[tag] = value
        self.tag_group.add_tag(tag)
    def validate(self):
        if len(self.new_tag_info)!=0 or len(self.remove_tag)!=0:
            title = 'Modify information confirmation'
            remove_info=";".join(self.remove_tag)
            add_info="\n".join([f"{k}={v}" for k,v in self.new_tag_info.items()])
            content = f"""You removed the following information from the structure: \n{remove_info}  \nadded the following information: \n{add_info}"""

            w = MessageBox(title, content, self)

            w.setClosableOnMaskClicked(True)


            if w.exec():

                return True
            else:
                return False
        return True

class ShiftEnergyMessageBox(MessageBoxBase):
    """Dialog for energy baseline shift parameters."""

    def __init__(self, parent=None, tip="Group regex patterns (comma separated)"):
        super().__init__(parent)
        self.titleLabel = CaptionLabel(tip, self)
        self.titleLabel.setWordWrap(True)
        self.groupEdit = LineEdit(self)

        self._frame = QFrame(self)
        self.frame_layout = QGridLayout(self._frame)
        self.frame_layout.setContentsMargins(0, 0, 0, 0)
        self.frame_layout.setSpacing(2)

        self.genSpinBox = SpinBox(self)
        self.genSpinBox.setMaximum(100000000)
        self.sizeSpinBox = SpinBox(self)
        self.sizeSpinBox.setMaximum(999999)
        self.tolSpinBox = DoubleSpinBox(self)
        self.tolSpinBox.setDecimals(10)
        self.tolSpinBox.setMinimum(0)
        self.modeCombo = ComboBox(self)
        self.modeCombo.addItems([
            "REF_GROUP",
            "ZERO_BASELINE",
            "DFT_TO_NEP",
        ])
        self.modeCombo.setCurrentText("DFT_TO_NEP")


        self.frame_layout.addWidget(CaptionLabel("Max generations", self), 0, 0)
        self.frame_layout.addWidget(self.genSpinBox, 0, 1)
        self.frame_layout.addWidget(CaptionLabel("Population size", self), 1, 0)
        self.frame_layout.addWidget(self.sizeSpinBox, 1, 1)
        self.frame_layout.addWidget(CaptionLabel("Convergence tol", self), 2, 0)
        self.frame_layout.addWidget(self.tolSpinBox, 2, 1)
        self.frame_layout.addWidget(HyperlinkLabel(QUrl("https://github.com/brucefan1983/GPUMD/tree/master/tools/Analysis_and_Processing/energy-reference-aligner"),
                                                   "Alignment mode", self), 3, 0)
        self.frame_layout.addWidget(self.modeCombo, 3, 1)


        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self.groupEdit)
        self.viewLayout.addWidget(self._frame)

        self.yesButton.setText('Ok')
        self.cancelButton.setText('Cancel')
        self.widget.setMinimumWidth(250)




class ProgressDialog(FramelessDialog):
    """进度条弹窗"""
    def __init__(self,parent=None,title=""):
        pass
        super().__init__(parent)
        self.setStyleSheet('ProgressDialog{background:white}')


        FluentStyleSheet.DIALOG.apply(self)


        self.setWindowTitle(title)
        self.setFixedSize(300,100)
        self.__layout = QVBoxLayout(self)
        self.__layout.setContentsMargins(0,0,0,0)
        self.progressBar = ProgressBar(self)
        self.progressBar.setRange(0,100)
        self.progressBar.setValue(0)
        self.__layout.addWidget(self.progressBar)
        self.setLayout(self.__layout)
        self.__thread = LoadingThread(self, show_tip=False)
        self.__thread.finished.connect(self.close)

        self.__thread.progressSignal.connect(self.progressBar.setValue)
    def closeEvent(self,event):
        if self.__thread.isRunning():
            self.__thread.stop_work()
    def run_task(self,task_function,*args,**kwargs):
        self.__thread.start_work(task_function, *args, **kwargs)


class PeriodicTableDialog(FramelessDialog):
    """Dialog showing a simple periodic table."""

    elementSelected = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitleBar(FluentTitleBar(self))
        self.setWindowTitle("Periodic Table")
        self.setWindowIcon(QIcon(':/images/src/images/logo.svg'))
        self.resize(400, 350)


        with open(os.path.join(module_path, "Config/ptable.json"), "r", encoding="utf-8") as f:
            self.table_data = {int(k): v for k, v in json.load(f).items()}

        self.group_colors = {}
        for info in self.table_data.values():
            g = info.get("group", 0)
            if g not in self.group_colors:
                self.group_colors[g] = info.get("color", "#FFFFFF")

        self.__layout = QGridLayout(self)
        self.__layout.setContentsMargins(2, 2,2, 2)
        self.__layout.setSpacing(1)
        self.setLayout(self.__layout)
        self.__layout.setMenuBar(self.titleBar)

        # self.__layout.addWidget(self.titleBar,0,0,1,18)
        for num in range(1, 119):
            info = self.table_data.get(num)
            if not info:
                continue
            group = info.get("group", 0)
            period = self._get_period(num)
            row, col = self._grid_position(num, group, period)
            btn = QPushButton(info["symbol"], self)
            btn.setFixedSize(30,30)
            btn.setStyleSheet(f'background-color: {info.get("color", "#FFFFFF")};')
            btn.clicked.connect(lambda _=False, sym=info["symbol"]: self.elementSelected.emit(sym))
            self.__layout.addWidget(btn, row+1, col)
    def _get_period(self, num: int) -> int:
        if num <= 2:
            return 1
        elif num <= 10:
            return 2
        elif num <= 18:
            return 3
        elif num <= 36:
            return 4
        elif num <= 54:
            return 5
        elif num <= 86:
            return 6
        else:
            return 7

    def _grid_position(self, num: int, group: int, period: int) -> tuple[int, int]:
        if group == 0:
            if 57 <= num <= 71:
                row = 8
                col = num - 53
            elif 89 <= num <= 103:
                row = 9
                col = num - 85
            else:
                row, col = period, 1
        else:
            row, col = period, group
        return row - 1, col - 1



class DFTD3MessageBox(MessageBoxBase):
    """Dialog for DFTD3 parameters."""

    def __init__(self, parent=None, tip="DFTD3 correction"):
        super().__init__(parent)
        self.titleLabel = CaptionLabel(tip, self)
        self.titleLabel.setWordWrap(True)
        self.functionEdit = EditableComboBox(self)
        self.functionEdit.setPlaceholderText("dft d3 functional")
        functionals = [
            "b1b95",
            "b2gpplyp",
            "b2plyp",
            "b3lyp",
            "b3pw91",
            "b97d",
            "bhlyp",
            "blyp",
            "bmk",
            "bop",
            "bp86",
            "bpbe",
            "camb3lyp",
            "dsdblyp",
            "hcth120",
            "hf",
            "hse-hjs",
            "lc-wpbe08",
            "lcwpbe",
            "m11",
            "mn12l",
            "mn12sx",
            "mpw1b95",
            "mpwb1k",
            "mpwlyp",
            "n12sx",
            "olyp",
            "opbe",
            "otpss",
            "pbe",
            "pbe0",
            "pbe38",
            "pbesol",
            "ptpss",
            "pw6b95",
            "pwb6k",
            "pwpb95",
            "revpbe",
            "revpbe0",
            "revpbe38",
            "revssb",
            "rpbe",
            "rpw86pbe",
            "scan",
            "sogga11x",
            "ssb",
            "tpss",
            "tpss0",
            "tpssh",
            "b2kplyp",
            "dsd-pbep86",
            "b97m",
            "wb97x",
            "wb97m"
        ]
        self.functionEdit.addItems(functionals)
        self._frame = QFrame(self)
        self.frame_layout = QGridLayout(self._frame)
        self.frame_layout.setContentsMargins(0, 0, 0, 0)
        self.frame_layout.setSpacing(2)

        self.d1SpinBox = DoubleSpinBox(self)
        self.d1SpinBox.setMaximum(100000000)
        self.d1SpinBox.setDecimals(3)

        self.d1cnSpinBox = DoubleSpinBox(self)
        self.d1cnSpinBox.setMaximum(999999)


        self.modeCombo = ComboBox(self)
        self.modeCombo.addItems([
            # "NEP Only",
            # "DFT-D3 only",
            # "NEP with DFT-D3",
            "Add DFT-D3",
            "Subtract DFT-D3",

        ])
        self.modeCombo.setCurrentText("NEP Only")


        self.frame_layout.addWidget(CaptionLabel("D3 cutoff ", self), 0, 0)
        self.frame_layout.addWidget(self.d1SpinBox, 0, 1)
        self.frame_layout.addWidget(CaptionLabel("D3 cutoff _cn ", self), 1, 0)
        self.frame_layout.addWidget(self.d1cnSpinBox, 1, 1)

        self.frame_layout.addWidget(CaptionLabel("Alignment mode", self), 3, 0)
        self.frame_layout.addWidget(self.modeCombo, 3, 1)


        self.viewLayout.addWidget(self.titleLabel)
        self.viewLayout.addWidget(self.functionEdit)
        self.viewLayout.addWidget(self._frame)

        self.yesButton.setText('Ok')
        self.cancelButton.setText('Cancel')
        self.widget.setMinimumWidth(250)


    def validate(self):
        if self.modeCombo.currentIndex()!=0:
            if len(self.functionEdit.text()) == 0:

                self.functionEdit.setFocus()
                return False
        return True
class ProjectInfoMessageBox(MessageBoxBase):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._widget = QWidget(self)

        self.widget_layout = QGridLayout(self._widget)

        self.parent_combox=ComboBox(self._widget)
        self.project_name=LineEdit(self._widget)
        self.project_name.setPlaceholderText("The name of the project")

        self.project_note=TextEdit(self._widget)
        self.project_note.setMinimumSize(200,100)
        self.project_note.setPlaceholderText("Notes on the project")
        self.widget_layout.addWidget(CaptionLabel("Parent",self), 0, 0)

        self.widget_layout.addWidget(self.parent_combox, 0, 1)

        self.widget_layout.addWidget(CaptionLabel("Project Name",self), 1, 0)
        self.widget_layout.addWidget(self.project_name, 1, 1)
        self.widget_layout.addWidget(CaptionLabel("Project Note",self), 2, 0 )
        self.widget_layout.addWidget(self.project_note, 2, 1 )
        self.viewLayout.addWidget(self._widget)
    def validate(self):
        project_name=self.project_name.text().strip()
        if len(project_name)==0:
            return False
        return True



class ModelInfoMessageBox(MessageBoxBase):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)

        # ===== 根容器 =====
        self._widget = QWidget(self)
        self.viewLayout.addWidget(self._widget)
        root = QVBoxLayout(self._widget)
        root.setContentsMargins(0, 0, 0, 0)
        root.setSpacing(2)

        # ===== 顶部 Title =====
        titleBar = QFrame(self._widget)
        tLayout = QHBoxLayout(titleBar)
        tLayout.setContentsMargins(0, 0, 0, 0)
        tLayout.setSpacing(0)
        self.titleLabel = TitleLabel("Create / Edit Model", titleBar)

        self.titleLabel.setAlignment(Qt.AlignCenter)
        tLayout.addWidget(self.titleLabel)
        root.addWidget(titleBar)

        # ===== 基本信息（左） =====
        infoCard = QFrame(self._widget)
        info = QFormLayout(infoCard)
        info.setLabelAlignment(Qt.AlignRight)
        info.setHorizontalSpacing(5)
        info.setVerticalSpacing(2)

        self.parent_combox = ComboBox(infoCard)
        self.model_type_combox = ComboBox(infoCard)
        self.model_type_combox.addItems(["NEP"])
        self.model_name_edit = LineEdit(infoCard)
        self.model_name_edit.setPlaceholderText("The name of the model")

        info.addRow(CaptionLabel("Parent", self), self.parent_combox)
        info.addRow(CaptionLabel("Type", self), self.model_type_combox)
        info.addRow(CaptionLabel("Name", self), self.model_name_edit)

        # ===== RMSE（右）——就是 energy/force/virial 三个输入 =====
        rmseCard = QFrame(self._widget)
        rmse = QGridLayout(rmseCard)
        rmse.setContentsMargins(0, 0, 0, 0)
        rmse.setHorizontalSpacing(5)
        rmse.setVerticalSpacing(2)

        titleRmse = CaptionLabel("RMSE (energy / force / virial)", self)
        tf = titleRmse.font()
        tf.setBold(True)
        titleRmse.setFont(tf)

        self.energy_spinBox = LineEdit(rmseCard)
        self.force_spinBox  = LineEdit(rmseCard)
        self.virial_spinBox = LineEdit(rmseCard)
        self.energy_spinBox.setText("0")
        self.force_spinBox.setText("0")
        self.virial_spinBox.setText("0")


        validator = QDoubleValidator(bottom=-1e12, top=1e12, decimals=2)
        for w in (self.energy_spinBox, self.force_spinBox, self.virial_spinBox):
            w.setValidator(validator)
            w.setPlaceholderText("0.0")

        r = 0
        rmse.addWidget(titleRmse, r, 0, 1, 3)
        r += 1
        rmse.addWidget(CaptionLabel("energy", self), r, 0)
        rmse.addWidget(self.energy_spinBox, r, 1)
        rmse.addWidget(CaptionLabel("meV/atom", self), r, 2)
        r += 1
        rmse.addWidget(CaptionLabel("force",  self), r, 0)
        rmse.addWidget(self.force_spinBox,  r, 1)
        rmse.addWidget(CaptionLabel("meV/Å",    self), r, 2)
        r += 1
        rmse.addWidget(CaptionLabel("virial", self), r, 0)
        rmse.addWidget(self.virial_spinBox, r, 1)
        rmse.addWidget(CaptionLabel("meV/atom", self), r, 2)
        r += 1
        rmse.setColumnStretch(1, 1)

        # ===== 第一行：基本信息 + RMSE 并排 =====
        row1 = QHBoxLayout()
        row1.setContentsMargins(0, 0, 0, 0)
        row1.setSpacing(2)
        row1.addWidget(infoCard, 2)
        row1.addWidget(rmseCard, 1)
        root.addLayout(row1)

        # ===== 文件路径（整行） =====
        pathCard = QFrame(self._widget)
        path = QFormLayout(pathCard)
        path.setLabelAlignment(Qt.AlignRight)
        path.setHorizontalSpacing(5); path.setVerticalSpacing(3)


        structureRow = QWidget(pathCard)
        h = QHBoxLayout(structureRow)
        h.setContentsMargins(0, 0, 0, 0)
        h.setSpacing(3)
        self.train_path_edit = LineEdit(structureRow)
        self.train_path_edit.setPlaceholderText("model train path")
        self.train_path_edit.editingFinished.connect(self.check_path)
        browse = TransparentToolButton(FluentIcon.FOLDER_ADD, structureRow)
        browse.setFixedHeight(self.train_path_edit.sizeHint().height())
        browse.clicked.connect(self._pick_file)
        h.addWidget(self.train_path_edit, 1)
        h.addWidget(browse, 0)




        path.addRow(CaptionLabel("Path", self), structureRow)

        root.addWidget(pathCard)

        # ===== Tags（Notes 之前） =====
        tagsCard = QFrame(self._widget)
        tags = QFormLayout(tagsCard)
        tags.setLabelAlignment(Qt.AlignRight)
        tags.setHorizontalSpacing(0)
        tags.setVerticalSpacing(0)

        self.new_tag_edit = LineEdit(tagsCard)
        self.new_tag_edit.setPlaceholderText("Enter the tag and press Enter")
        self.new_tag_edit.returnPressed.connect(lambda :self.add_tag(self.new_tag_edit.text()))
        self.tag_group = TagGroup(parent=self)

        tags.addRow(CaptionLabel("Tags", self), self.new_tag_edit )
        tags.addRow(CaptionLabel(""), self.tag_group)  # 让 TagGroup 独占一行
        root.addWidget(tagsCard)

        # ===== Notes（最后） =====
        notesCard = QFrame(self._widget)
        notes = QFormLayout(notesCard)
        notes.setLabelAlignment(Qt.AlignRight)
        notes.setHorizontalSpacing(5)
        notes.setVerticalSpacing(0)

        self.model_note_edit = TextEdit(notesCard)
        self.model_note_edit.setPlaceholderText("Notes on the model")
        self.model_note_edit.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        # self.model_note_edit.setMinimumHeight(30)

        notes.addRow(CaptionLabel("Notes", self), self.model_note_edit)
        root.addWidget(notesCard)

        # 允许底部区域伸展
        root.addStretch(1)



    # 简单文件选择器
    def _pick_file(self):
        path=call_path_dialog(self,"Select the model folder path","directory")

        if path:
            self.train_path_edit.setText(path)
            self.check_path()
    def add_tag(self,tag ):
        if self.tag_group.has_tag(tag):
            MessageManager.send_info_message(f"{tag} already exists!")
            return

        self.tag_group.add_tag(tag)
    def check_path(self):
        _path=self.train_path_edit.text()
        path=Path(_path)
        if not path.exists():
            MessageManager.send_message_box(f"{_path} does not exist!")
            return
        if self.model_type_combox.currentText()=="NEP":
            model_file=path.joinpath("nep.txt")
            if not model_file.exists():
                MessageManager.send_message_box("No 'nep.txt' found in the specified path. Its presence is not strictly required, but please make sure you know what you are doing.")

            data_file=path.joinpath("train.xyz")
            if not data_file.exists():
                MessageManager.send_message_box("No 'nep.txt' found in the specified path. Its presence is not strictly required, but please make sure you know what you are doing.")
                # data_size=0
                energy=0
                force=0
                virial=0
            else:

                # data_size=get_xyz_nframe(data_file)
                # if data_size
                energy_array=read_nep_out_file(path.joinpath("energy_train.out"))
                energy = get_rmse(energy_array[:,0],energy_array[:,1])*1000
                force_array=read_nep_out_file(path.joinpath("force_train.out"))
                force = get_rmse(force_array[:,:3],force_array[:,3:])*1000
                virial_array=read_nep_out_file(path.joinpath("virial_train.out"))
                virial = get_rmse(virial_array[:,:6],virial_array[:,6:])*1000

            self.force_spinBox.setText(str(round(force,2)))
            self.energy_spinBox.setText(str(round(energy,2)))
            self.virial_spinBox.setText(str(round(virial,2)))
    def get_dict(self):
        path=Path(self.train_path_edit.text())
        data_file=path.joinpath("train.xyz")
        data_size = get_xyz_nframe(data_file)
        return dict(
            # project_id=self.,
            name=self.model_name_edit.text().strip(),
            model_type=self.model_type_combox.currentText(),
            model_path=self.train_path_edit.text().strip(),
            # model_file=path.joinpath("nep.txt"),
            # data_file=data_file,
            data_size=data_size,
            energy=float(self.energy_spinBox.text().strip()),
            force=float(self.force_spinBox.text().strip()),
            virial=float(self.virial_spinBox.text().strip()),

            notes=self.model_note_edit.toPlainText(),
            tags=list(self.tag_group.tags.keys()),
            parent_id=self.parent_combox.currentData()
        )
class AdvancedModelSearchDialog(MessageBoxBase):
    """
    仅负责收集搜索条件并发送信号，不执行查询。
    使用：
        dlg = AdvancedModelSearchDialog(parent)
        dlg.searchRequested.connect(handle_search_params_dict)
        dlg.show()
    """
    searchRequested = Signal(dict)  # 发出参数字典

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Advanced Search - Models")
        # self.setDraggable(True)
        self.setModal(False)
        # self.resize(640, 520)
        self._build_ui()
        self._wire_events()

    # ---------- UI ----------
    def _build_ui(self):
        root = QVBoxLayout()
        root.setContentsMargins(0, 0, 0, 0)
        root.setSpacing(3)
        self.viewLayout.addLayout(root)
        # Title
        titleBar = QFrame(self)
        tLay = QHBoxLayout(titleBar); tLay.setContentsMargins(0, 0, 0, 0)
        self.titleLabel = TitleLabel("Advanced Model Search", titleBar)
        # f = self.titleLabel.font(); f.setPointSize(f.pointSize() + 3); f.setBold(True)
        # self.titleLabel.setFont(f)
        self.titleLabel.setAlignment(Qt.AlignCenter)
        tLay.addWidget(self.titleLabel)
        root.addWidget(titleBar)

        # 表单
        formCard = QFrame(self); form = QFormLayout(formCard)
        form.setLabelAlignment(Qt.AlignRight); form.setHorizontalSpacing(3); form.setVerticalSpacing(3)

        # Project IDs（逗号分隔）
        self.projectIdsEdit = LineEdit(formCard)
        self.projectIdsEdit.setPlaceholderText("e.g. 1 or 1,3,5")
        self.includeDescendantsChk = CheckBox("Include sub-projects", formCard)
        self.includeDescendantsChk.setChecked(True)

        # Parent id
        self.parentIdEdit = LineEdit(formCard)
        self.parentIdEdit.setPlaceholderText("None or integer")
        self.parentIdEdit.setValidator(QIntValidator())

        # 模糊
        self.nameContainsEdit = LineEdit(formCard)
        self.nameContainsEdit.setPlaceholderText("contains in name")
        self.notesContainsEdit = LineEdit(formCard)
        self.notesContainsEdit.setPlaceholderText("contains in notes")

        # 类型
        self.modelTypeCombo = ComboBox(formCard)
        self.modelTypeCombo.addItems(["<Any>", "NEP", "DeepMD", "Other"])

        # 标签
        self.tagsAllEdit  = LineEdit(formCard); self.tagsAllEdit.setPlaceholderText("tag1, tag2 (AND)")
        self.tagsAnyEdit  = LineEdit(formCard); self.tagsAnyEdit.setPlaceholderText("tag1, tag2 (OR)")
        self.tagsNoneEdit = LineEdit(formCard); self.tagsNoneEdit.setPlaceholderText("tag1, tag2 (NOT)")

        # 排序与分页
        self.orderAscChk = CheckBox("Order by created_at ascending", formCard)
        self.orderAscChk.setChecked(True)
        self.limitEdit  = LineEdit(formCard); self.limitEdit.setPlaceholderText("e.g. 100"); self.limitEdit.setValidator(QIntValidator(0, 10**9))
        self.offsetEdit = LineEdit(formCard); self.offsetEdit.setPlaceholderText("e.g. 0");   self.offsetEdit.setValidator(QIntValidator(0, 10**9))

        # 加入表单
        form.addRow(CaptionLabel("Project ID(s):",self), self.projectIdsEdit)
        form.addRow(CaptionLabel("",self), self.includeDescendantsChk)
        form.addRow(CaptionLabel("Parent ID:",self), self.parentIdEdit)
        form.addRow(CaptionLabel("Model Type:",self), self.modelTypeCombo)
        form.addRow(CaptionLabel("Name contains:",self), self.nameContainsEdit)
        form.addRow(CaptionLabel("Notes contains:",self), self.notesContainsEdit)
        form.addRow(CaptionLabel("Tags (ALL):",self), self.tagsAllEdit)
        form.addRow(CaptionLabel("Tags (ANY):",self), self.tagsAnyEdit)
        form.addRow(CaptionLabel("Tags (NOT):",self), self.tagsNoneEdit)
        form.addRow(CaptionLabel("Order:",self), self.orderAscChk)
        form.addRow(CaptionLabel("Limit:",self), self.limitEdit)
        form.addRow(CaptionLabel("Offset:",self), self.offsetEdit)

        root.addWidget(formCard)

        # 按钮区

        self.buttonLayout.removeWidget(self.yesButton)
        self.buttonLayout.removeWidget(self.cancelButton)
        self.yesButton.hide()
        self.cancelButton.hide()
        self.searchBtn = PrimaryPushButton("Search", self)
        self.resetBtn  = PrimaryPushButton("Reset", self)
        self.closeBtn  = PrimaryPushButton("Close", self)
        self.buttonLayout.addWidget(self.searchBtn)
        self.buttonLayout.addWidget(self.resetBtn)
        self.buttonLayout.addWidget(self.closeBtn)


        root.addStretch(1)


    # ---------- 事件 ----------
    def _wire_events(self):
        self.searchBtn.clicked.connect(self._emit_params)
        self.resetBtn.clicked.connect(self._on_reset)
        self.closeBtn.clicked.connect(self.reject)
        # Enter 键触发搜索
        self.projectIdsEdit.returnPressed.connect(self._emit_params)
        self.nameContainsEdit.returnPressed.connect(self._emit_params)
        self.notesContainsEdit.returnPressed.connect(self._emit_params)
        self.tagsAllEdit.returnPressed.connect(self._emit_params)
        self.tagsAnyEdit.returnPressed.connect(self._emit_params)
        self.tagsNoneEdit.returnPressed.connect(self._emit_params)

    # ---------- 参数构造 ----------
    @staticmethod
    def _split_csv(text: str) -> list[str]:
        if not text:
            return []
        out, seen = [], set()
        for part in text.split(","):
            s = part.strip()
            if not s:
                continue
            key = s.lower()
            if key in seen:
                continue
            seen.add(key)
            out.append(s)
        return out

    @staticmethod
    def _parse_project_ids(text: str) -> list[int]:
        if not text.strip():
            return []
        ids = []
        for part in text.split(","):
            p = part.strip()
            if not p:
                continue
            try:
                ids.append(int(p))
            except ValueError:
                pass
        return ids

    def build_params(self) -> Dict[str, Any]:
        """收集并返回与 search_models_advanced 对应的参数字典。"""
        project_ids = self._parse_project_ids(self.projectIdsEdit.text())
        mt_text = self.modelTypeCombo.currentText()
        model_type = None if mt_text == "<Any>" else mt_text

        parent_text = self.parentIdEdit.text().strip()
        parent_id_val = int(parent_text) if parent_text.isdigit() else None

        params: Dict[str, Any] = dict(
            project_id=(
                project_ids[0] if len(project_ids) == 1
                else (project_ids if project_ids else None)
            ),
            include_descendants=self.includeDescendantsChk.isChecked(),
            parent_id=parent_id_val,
            name_contains=(self.nameContainsEdit.text().strip() or None),
            notes_contains=(self.notesContainsEdit.text().strip() or None),
            model_type=model_type,
            tags_all=self._split_csv(self.tagsAllEdit.text()),
            tags_any=self._split_csv(self.tagsAnyEdit.text()),
            tags_none=self._split_csv(self.tagsNoneEdit.text()),
            order_by_created_asc=self.orderAscChk.isChecked(),
        )

        limit_text = self.limitEdit.text().strip()
        if limit_text:
            params["limit"] = int(limit_text)
        offset_text = self.offsetEdit.text().strip()
        if offset_text:
            params["offset"] = int(offset_text)

        return params

    # ---------- 对外：发出信号 ----------
    def _emit_params(self):
        params = self.build_params()
        self.searchRequested.emit(params)

    # ---------- 重置 ----------
    def _on_reset(self):
        self.projectIdsEdit.clear()
        self.includeDescendantsChk.setChecked(True)
        self.parentIdEdit.clear()
        self.modelTypeCombo.setCurrentIndex(0)
        self.nameContainsEdit.clear()
        self.notesContainsEdit.clear()
        self.tagsAllEdit.clear()
        self.tagsAnyEdit.clear()
        self.tagsNoneEdit.clear()
        self.orderAscChk.setChecked(True)
        self.limitEdit.clear()
        self.offsetEdit.clear()


class TagEditDialog(MessageBoxBase):
    """Dialog for editing tag properties."""

    def __init__(self, name: str, color: str, notes: str, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Edit Tag")
        # self.resize(300, 200)

        layout = QVBoxLayout()
        self.viewLayout.addLayout(layout)
        form = QFormLayout()
        self.nameEdit = LineEdit(self)
        self.nameEdit.setText(name)
        self.colorEdit = LineEdit(self)
        self.colorEdit.setText(color)
        self.colorBtn = PrimaryPushButton("...", self)
        self.colorBtn.setFixedWidth(30)
        colorLayout = QHBoxLayout()
        colorLayout.setContentsMargins(0, 0, 0, 0)
        colorLayout.setSpacing(3)
        colorLayout.addWidget(self.colorEdit)
        colorLayout.addWidget(self.colorBtn)
        colorWidget = QWidget(self)
        colorWidget.setLayout(colorLayout)
        self.notesEdit = TextEdit(self)
        self.notesEdit.setPlainText(notes)

        form.addRow("Name", self.nameEdit)
        form.addRow("Color", colorWidget)
        form.addRow("Notes", self.notesEdit)
        layout.addLayout(form)



        self.colorBtn.clicked.connect(self._choose_color)


    def _choose_color(self):
        color_dialog = ColorDialog(QColor(self.colorEdit.text()),"Edit Tag Color", self)
        if color_dialog.exec():
            self.colorEdit.setText(color_dialog.color.name())

    def get_values(self) -> tuple[str, str, str]:
        return (
            self.nameEdit.text().strip(),
            self.colorEdit.text().strip(),
            self.notesEdit.toPlainText().strip(),
        )

class TagManageDialog(MessageBoxBase):
    """Dialog to create, edit and remove tags."""

    def __init__(self, tag_service, parent=None):
        super().__init__(parent)
        self._parent=parent
        self.tag_changed=False
        self.setWindowTitle("Manage Tags")
        self.tag_service = tag_service
        self._tag_map: dict[str, int] = {}
        # self.resize(360, 240)

        self._layout = QVBoxLayout()
        self.new_tag_edit = LineEdit(self)
        self.new_tag_edit.setMinimumWidth(300)
        self.new_tag_edit.setPlaceholderText("Enter the tag and press Enter")
        self.new_tag_edit.returnPressed.connect(self.add_tag)
        self.tag_group = TagGroup(parent=self)
        self.tag_group.setMinimumHeight(100)
        self.tag_group.tagRemovedSignal.connect(self.remove_tag)
        self._layout.addWidget(self.new_tag_edit)
        self._layout.addWidget(self.tag_group)
        self.viewLayout.addLayout(self._layout)


        self._load_tags()

    def _load_tags(self):
        for tag in self.tag_service.get_tags():
            btn = self.tag_group.add_tag(tag.name, color=tag.color)
            btn.setToolTip(tag.notes)
            btn.installEventFilter(self)
            self._tag_map[tag.name] = tag.tag_id

    def add_tag(self):
        name = self.new_tag_edit.text().strip()
        if not name:
            return
        if self.tag_group.has_tag(name):
            MessageManager.send_info_message(f"{name} already exists!")
            return
        item = self.tag_service.create_tag(name)
        if item:
            btn = self.tag_group.add_tag(item.name, color=item.color)
            btn.setToolTip(item.notes)
            btn.installEventFilter(self)
            self._tag_map[item.name] = item.tag_id
        self.new_tag_edit.clear()

    def remove_tag(self, name: str):
        tag_id = self._tag_map.pop(name, None)
        if tag_id is not None:
            self.tag_service.remove_tag(tag_id)

    def eventFilter(self, obj, event):

        if isinstance(obj, TagPushButton) and event.type() == QEvent.ContextMenu:
            old_name = obj.text()
            tag_id = self._tag_map.get(old_name)
            dlg = TagEditDialog(old_name, obj.backgroundColor, obj.toolTip(), self._parent)
            if dlg.exec():
                new_name, color, notes = dlg.get_values()
                if not new_name:
                    return True
                if new_name != old_name and self.tag_group.has_tag(new_name):
                    MessageManager.send_info_message(f"{new_name} already exists!")
                    return True
                self.tag_changed=True
                self.tag_service.update_tag(tag_id, name=new_name, color=color, notes=notes)
                obj.setText(new_name)
                obj.setBackgroundColor(color)
                obj.setToolTip(notes)
                if new_name != old_name:
                    self.tag_group.tags[new_name] = self.tag_group.tags.pop(old_name)
                    self._tag_map[new_name] = self._tag_map.pop(old_name)
            return True
        return super().eventFilter(obj, event)



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2025/8/31 08:49
# @Author  : 兵
# @email    : 1747193328@qq.com


from PySide6.QtCore import QObject, QTimer, Qt, QPoint, Signal
from PySide6.QtGui import QCursor, QIcon
from PySide6.QtWidgets import QWidget, QVBoxLayout,QGroupBox
from qfluentwidgets import RoundMenu, Action, TreeView, TableView, MessageBox, FluentIcon, SearchLineEdit

from NepTrainKit.core import MessageManager
from NepTrainKit.core.dataset import DatasetManager

from NepTrainKit.core.dataset.database import Database
from NepTrainKit.core.dataset.services import ModelService, ProjectService, ProjectItem
from NepTrainKit.custom_widget import IdNameTableModel, TreeModel, TreeItem, ProjectInfoMessageBox, TagManageDialog
from NepTrainKit.views import KitToolBarBase


class ProjectWidget(QWidget,DatasetManager):
    project_item_dict={}
    projectChangedSignal=Signal(ProjectItem)
    def __init__(self, parent=None):
        super(ProjectWidget, self).__init__(parent)
        self._parent=parent


        self._view = TreeView()
        self._view.clicked.connect(self.item_clicked)
        self._view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

        # self._view.setIndentation(10)
        self._view.header().setDefaultSectionSize(5)
        self._view.header().setStretchLastSection(True)
        self._model=TreeModel()

        self._view.setModel(self._model)
        self._model.setHeader(["(ID) Project Name","ID",""])
        # self._model.count_column=2


        self._view.setColumnHidden(1,True)
        self._view.setColumnWidth(0,150)

        self._layout = QVBoxLayout(self)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0,0,0,0)
        self._layout.addWidget(self._view)
        self.create_menu()

        QTimer.singleShot(1, self.load)


    def item_clicked(self,index):
        item = index.internalPointer()
        project=self.project_item_dict[item.data(1)]
        self.projectChangedSignal.emit(project)
    def create_menu(self):
        self._menu_pos = QPoint()
        self.menu = RoundMenu(parent=self)

        create_action = Action( "New",self.menu)
        create_action.triggered.connect(lambda :self.create_project(modify=False))
        self.menu.addAction(create_action)
        modify_action = Action( "Modify",self.menu)
        modify_action.triggered.connect(lambda :self.create_project(modify=True))
        self.menu.addAction(modify_action)
        delete_action = Action( "Delete", self.menu)
        delete_action.triggered.connect(self.remove_project)
        self.menu.addAction(delete_action)


        self._view.customContextMenuRequested.connect(self.show_menu)
    def show_menu(self,pos):
        self._menu_pos=pos
        self.menu.exec_(self.mapToGlobal(pos))
    def create_project(self,modify=False):
        box = ProjectInfoMessageBox(self._parent)
        index=self._view.indexAt(self._menu_pos)
        box.parent_combox.addItem("Top Project",userData=None)
        for  project in self.project_item_dict.values():
            box.parent_combox.addItem(project.name, userData=project.project_id)

        if index.row()!=-1:
            item=index.internalPointer()
            project_id=item.data(1)
            project=self.project_item_dict[project_id]

            box.parent_combox.setCurrentText(project.name)
        else:
            box.parent_combox.setCurrentText("Top Project")
            if modify:
                return
            project_id=None
        box.setWindowTitle(f"Project Info")
        if modify:
            current_project = self.project_item_dict[project_id]
            box.project_name.setText(current_project.name)
            box.project_note.setText(current_project.notes)


            if current_project.parent_id is not None:
                parent_project = self.project_item_dict[current_project.parent_id]
                box.parent_combox.setCurrentText(parent_project.name)
            else:
                box.parent_combox.setCurrentText("Top Project")

        if not box.exec_():
            return
        name=box.project_name.text().strip()
        note=box.project_note.toPlainText().strip()
        project_id=box.parent_combox.currentData()
        if modify:
            self.project_service.modify_project(current_project.project_id,
                                                name=name,notes=note,parent_id=project_id)
            self.load_all_projects()

            MessageManager.send_success_message("Project modification successful")
            return

        project = self.project_service.create_project(
            name=name,
            notes=note,
            parent_id=project_id,

        )
        if project is None:
            MessageManager.send_error_message("Failed to create project")
        else:
            MessageManager.send_success_message("Project created successfully")
            self.load_all_projects()


    def remove_project(self):


        index = self._view.indexAt(self._menu_pos)

        if index.row() == -1:
            return

        item = index.internalPointer()

        project_id = item.data(1)
        box = MessageBox("Ask",
                         "Do you want to delete this item?\nIf you delete it, all items under it will be deleted!",
                         self._parent)
        box.exec_()
        if box.result() == 0:
            return

        self.project_service.remove_project(project_id=project_id)

        MessageManager.send_success_message("Project deleted successfully")
        self.load_all_projects()

    def load(self):
        self.load_all_projects()


    def _build_tree(self,project,parent:TreeItem):
        child = TreeItem((f"({project.project_id}){project.name}", project.project_id,project.model_num))
        child.icon=FluentIcon.FOLDER.icon()

        self.project_item_dict[project.project_id] = project
        parent.appendChild(child)
        for item in project.children:
            self._build_tree(item,child)
        return child



    def load_all_projects(self):
        self._model.clear()
        all_project = self.project_service.search_projects(parent_id=None)
        self._model.beginResetModel()
        for project in all_project:

            self._build_tree(project,self._model.rootItem)


        self._model.endResetModel()



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2025/8/31 09:54
# @Author  : 兵
# @email    : 1747193328@qq.com
from PySide6.QtCore import QAbstractItemModel, QModelIndex, Qt, QRect, QSize


from PySide6.QtGui import QColor, QFont, QPainter, QPen, QBrush,QIcon
from PySide6.QtWidgets import (
    QStyledItemDelegate, QStyleOptionViewItem, QWidget,
    QLineEdit, QCompleter
)

class TreeItem:
    __slots__ = ("parentItem", "childItems", "itemData", "_row","icon")

    def __init__(self, data, parent=None):
        # data 建议用 tuple，一次性创建后不再改动，节省内存 & 更快
        self.parentItem = parent
        self.childItems = []
        self.icon:QIcon|None=None
        self.itemData = tuple(data) if not isinstance(data, tuple) else data
        self._row = -1  # 本节点在父节点中的行号（O(1) 获取）

    # ------- 结构维护：确保父子关系与 _row 始终一致 -------
    def appendChild(self, child):
        # ⚡ 必须维护 parent & row
        child.parentItem = self
        child._row = len(self.childItems)
        self.childItems.append(child)
        return child
    def insertChild(self, row, child):
        # 安全行号
        if row < 0:
            row = 0
        elif row > len(self.childItems):
            row = len(self.childItems)
        child.parentItem = self
        child._row = row
        self.childItems.insert(row, child)
        # ⚡ 更新后续兄弟的 _row（一次性顺序修正，仍是 O(n) 但仅在插入/删除发生时）
        for i in range(row + 1, len(self.childItems)):
            self.childItems[i]._row = i

    def removeChild(self, row):
        if 0 <= row < len(self.childItems):
            # ⚡ 不用 list.remove(child) —— 直接按行号删除（避免 O(n) 查找）
            self.childItems.pop(row)
            # ⚡ 批量修正行号
            for i in range(row, len(self.childItems)):
                self.childItems[i]._row = i

    def clear(self):
        # 大量删除时，更快
        self.childItems.clear()

    # ------- 快速访问 -------
    def child(self, row):
        # ⚡ 边界检查，安全且更快
        if 0 <= row < len(self.childItems):
            return self.childItems[row]
        return None

    def childCount(self):
        return len(self.childItems)

    def columnCount(self):
        return len(self.itemData)

    def data(self, column):

        # ⚡ 边界检查
        if 0 <= column < len(self.itemData):
            return self.itemData[column]
        return None

    def setRow(self, row: int):
        self._row = row

    def row(self):
        return self._row

    def parent(self):
        return self.parentItem


class TreeModel(QAbstractItemModel):
    def __init__(self, parent=None):
        super(TreeModel, self).__init__(parent)
        self.rootItem = TreeItem(())  # 根的 header 在 setHeader 设置
        # 预缓存：flags 常量
        self.count_column = None
        self._base_flags = super().flags(QModelIndex())

    # ------- 数据访问 -------
    def data(self, index: QModelIndex, role: int = Qt.DisplayRole):
        if not index.isValid():
            return None
        # ⚡ 快速路径：只处理显示文本，其他 role 直接返回 None
        item = index.internalPointer()
        col = index.column()

        if role==Qt.ItemDataRole.DecorationRole:

            if col==0:
                return item.icon


        if role != Qt.DisplayRole  :


            return None



        if col == self.count_column:
            return item.childCount()
        else:
            # 原实现为 item.data(0)
            return item.data(col)



    def flags(self, index: QModelIndex):
        if not index.isValid():
            return self._base_flags | Qt.ItemIsEnabled
        # ⚡ 避免 super().flags(index) 的重复调用（会创建临时对象）
        return self._base_flags | Qt.ItemIsEnabled | Qt.ItemIsSelectable

    # ------- 表头 -------
    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.rootItem.data(section)
        return None

    # ------- 索引构造 -------
    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex()) -> QModelIndex:
        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        parentItem = self.rootItem if not parent.isValid() else parent.internalPointer()
        childItem = parentItem.child(row)
        if childItem is None:
            return QModelIndex()
        # ⚡ 使用内部指针保存 item，row/column 也写入（Qt 会缓存一部分）
        return self.createIndex(row, column, childItem)

    def parent(self, index: QModelIndex) -> QModelIndex:
        if not index.isValid():
            return QModelIndex()
        childItem = index.internalPointer()
        parentItem = childItem.parent()
        if parentItem is None or parentItem == self.rootItem:
            return QModelIndex()
        # ⚡ 这里依赖子节点维护好的 _row，O(1)
        return self.createIndex(parentItem.row(), 0, parentItem)

    # ------- 维度 -------
    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        parentItem = self.rootItem if not parent.isValid() else parent.internalPointer()
        return parentItem.childCount()

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        # ⚡ 表头长度固定 O(1)
        return self.rootItem.columnCount()
    def clear(self):
        self.rootItem.clear()

    # ------- 批量重建 / 表头 -------
    def setHeader(self, header):
        # ⚡ 一次性重置模型（大量变更时显著更快）
        self.beginResetModel()
        self.rootItem = TreeItem(tuple(header))
        self.endResetModel()

    # ------- 可选：批量构建接口（推荐用于大数据载入） -------
    def rebuild(self, builder_fn):
        """
        builder_fn(rootItem: TreeItem) -> None
        外部传入一个函数，在里面把整棵树构建到 rootItem 下。
        好处：一次 beginResetModel/endResetModel，大幅减少 UI 刷新和 QModelIndex 生成。
        """
        self.beginResetModel()
        # 清空旧树
        self.rootItem.clear()
        # 构建新树
        builder_fn(self.rootItem)
        self.endResetModel()

    # ------- 可选：插入/删除行（若你需要动态更新 UI） -------
    def insertRows(self, row, count, parent=QModelIndex()):
        parentItem = self.rootItem if not parent.isValid() else parent.internalPointer()
        if row < 0 or row > parentItem.childCount() or count <= 0:
            return False
        self.beginInsertRows(parent, row, row + count - 1)
        for i in range(count):
            parentItem.insertChild(row + i, TreeItem((), parentItem))
        self.endInsertRows()
        return True

    def removeRows(self, row, count, parent=QModelIndex()):
        parentItem = self.rootItem if not parent.isValid() else parent.internalPointer()
        if row < 0 or row + count > parentItem.childCount() or count <= 0:
            return False
        self.beginRemoveRows(parent, row, row + count - 1)
        for _ in range(count):
            parentItem.removeChild(row)
        self.endRemoveRows()
        return True

    def add_item(self,parent,site):

        self.beginResetModel()
        if   isinstance(parent,int):
            if parent==-1:
                parent = self.rootItem
            else:
                parent= self.rootItem.child(parent)
        elif isinstance(parent, TreeItem):
            pass
        else:
            raise ValueError("parent must be TreeItem or int")
        child = parent.appendChild(TreeItem(site, parent))

        self.endResetModel()
        return child




class TagDelegate(QStyledItemDelegate):
    """
    一个简单的 Tag 代理：
    - 在 TreeView 中以彩色小圆角矩形显示标签
    - 双击可编辑，输入逗号分隔的标签
    - 可选支持自动补全（传入 tag_list）
    """

    def __init__(self, parent=None, tag_list=None):
        super().__init__(parent)
        self.tag_list = tag_list or []   # 用于 QCompleter
    def sizeHint(self, option, index):
        tags = index.data(Qt.DisplayRole)
        if not tags:
            return super().sizeHint(option, index)

        if isinstance(tags, str):
            tags = [t.strip() for t in tags.split(",") if t.strip()]

        font = option.font
        font.setPointSize(font.pointSize() - 1)
        metrics = option.fontMetrics if option.fontMetrics else option.widget.fontMetrics()

        spacing = 4
        padding = 6
        height = metrics.height() + 8

        # 计算总宽度
        total_width = 4  # 左侧内边距
        for tag_info in tags:
            tag=tag_info["name"]
            tag_width = metrics.horizontalAdvance(tag) + padding * 2
            total_width += tag_width + spacing

        return QSize(total_width, height)

    # --- 显示 ---
    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index):
        # 从 model 拿到数据，假设是 list[str]
        tags = index.data(Qt.DisplayRole)
        if not tags:
            super().paint(painter, option, index)
            return



        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)

        rect = option.rect
        x, y = rect.x() + 4, rect.y() + 2
        spacing = 4
        padding = 6
        tag_height = rect.height() - 4

        font = option.font
        font.setPointSize(font.pointSize() - 1)
        painter.setFont(font)

        for tag_info in tags:
            tag=tag_info["name"]
            color = tag_info["color"]
            tag_width = painter.fontMetrics().horizontalAdvance(tag) + padding * 2
            r = QRect(x, y, tag_width, tag_height)

            # 背景
            painter.setBrush(QBrush(QColor(color)))
            painter.setPen(Qt.NoPen)
            painter.drawRoundedRect(r, 6, 6)

            # 文本
            painter.setPen(QPen(Qt.black))
            painter.drawText(r, Qt.AlignCenter, tag)

            x += tag_width + spacing
            if x > rect.right():
                break  # 超出就不画

        painter.restore()

    # --- 编辑器 ---
    def createEditor(self, parent: QWidget, option, index):
        editor = QLineEdit(parent)
        editor.setPlaceholderText("逗号分隔多个标签")
        if self.tag_list:
            completer = QCompleter(self.tag_list, editor)
            completer.setCaseSensitivity(Qt.CaseInsensitive)
            editor.setCompleter(completer)
        return editor

    def setEditorData(self, editor: QLineEdit, index):
        tags = index.data(Qt.EditRole)
        if isinstance(tags, list):
            editor.setText(", ".join(tags))
        elif isinstance(tags, str):
            editor.setText(tags)

    def setModelData(self, editor: QLineEdit, model, index):
        text = editor.text().strip()
        tags = [t.strip() for t in text.split(",") if t.strip()]
        model.setData(index, tags, Qt.EditRole)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import os
import traceback
from dataclasses import dataclass
from typing import Callable, Optional, Protocol
from loguru import logger
import importlib
from .utils import get_nep_type
from NepTrainKit import get_user_config_path, utils


class ResultDataProtocol(Protocol):
    load_flag: bool
    @classmethod
    def from_path(cls, path: str, *args, **kwargs):
        ...


class ResultLoader(Protocol):
    name: str
    def matches(self, path: str) -> bool: ...
    def load(self, path: str): ...


_RESULT_LOADERS: list[ResultLoader] = []


def register_result_loader(loader: ResultLoader):
    _RESULT_LOADERS.append(loader)
    return loader


def matches_result_loader(path)->bool:
    for loader in _RESULT_LOADERS:
        try:
            if loader.matches(path):
                return True
        except:
            pass
    return False

def load_result_data(path: str):
    for loader in _RESULT_LOADERS:
        try:
            if loader.matches(path):

                return loader.load(path)
        except Exception:
            # Fail soft per loader
            logger.debug(f"{loader.name}failed to load {path}")
            continue

    # Fallback: try format importers to convert to EXTXYZ and load as NEP dataset
    # try:
    #     structures = import_structures(path)
    #     if structures:
    #         base = os.path.basename(path.rstrip("/\\"))
    #         base_noext = os.path.splitext(base)[0]
    #         target_dir = os.path.join(get_user_config_path(), "imported")
    #         os.makedirs(target_dir, exist_ok=True)
    #         target_xyz = os.path.join(target_dir, f"{base_noext}.xyz")
    #         write_extxyz(target_xyz, structures)
    #         return nep.NepTrainResultData.from_path(target_xyz)
    # except Exception:
    #     pass
    return None


class DeepmdFolderLoader:
    name = "deepmd_folder"
    def matches(self, path: str) -> bool:
        if not os.path.isdir(path):
            return False
        try:
            mod = importlib.import_module('.deepmd', __package__)
            return mod.is_deepmd_path(path)
        except Exception:
            return False
    def load(self, path: str):
        mod = importlib.import_module('.deepmd', __package__)
        return mod.DeepmdResultData.from_path(path)



class NepModelTypeLoader:
    def __init__(self, name: str, model_types: set[int], factory_path: str):
        self.name = name
        self._types = set(model_types)
        self._factory_path = factory_path
        self._factory = None
        self.model_type=None

    def matches(self, path: str) -> bool:
        if os.path.isdir(path):
            return False
        dir_path = os.path.dirname(path)
        self.model_type = get_nep_type(os.path.join(dir_path, "nep.txt"))
        return self.model_type in self._types and path.endswith(".xyz")

    def load(self, path: str):
        # Resolve factory lazily
        if self._factory is None:
            module_name, cls_name = self._factory_path.split(':', 1)
            mod = importlib.import_module(module_name)
            self._factory = getattr(mod, cls_name)
        # Pass through model_type for NepTrainResultData to keep behavior parity
        if self._factory_path.endswith(':NepTrainResultData'):
            return self._factory.from_path(path, model_type=self.model_type)
        return self._factory.from_path(path)



class OtherLoader:
    def matches(self,path: str) -> bool:
        try:
            imp_mod = importlib.import_module('.importers', __package__)
            return imp_mod.is_parseable(path)
        except Exception:
            return False

    def load(self, path: str):
        # Defer heavy parsing to the worker thread via ResultData.load_structures

        nep_mod = importlib.import_module('.nep', __package__)
        inst = nep_mod.NepTrainResultData.from_path(path)
        # If this is a LAMMPS dump, optionally prompt for element map

        imp_mod = importlib.import_module('.importers', __package__)
        lmp_imp = getattr(imp_mod, 'LammpsDumpImporter', None)
        if lmp_imp is not None and lmp_imp().matches(path):
            # Prompt user for an optional element list mapping type 1..N

            from PySide6.QtWidgets import QInputDialog
            prompt = "Please enter a list of elements (corresponding to type 1..N), separated by commas or spaces. \nFor example: Si O or Si,O"
            text, ok = QInputDialog.getText(None, "Element Mapping", prompt)

            if ok and text:
                raw = [t.strip() for t in str(text).replace(',', ' ').split() if t.strip()]
                if raw:
                    element_map = {i + 1: raw[i] for i in range(len(raw))}
                    # Attach to importer options so worker thread uses it
                    setattr(inst, '_import_options', {**getattr(inst, '_import_options', {}), 'element_map': element_map})
            else:
                return None

        return inst



# Register defaults immediately

register_result_loader(DeepmdFolderLoader())

register_result_loader(NepModelTypeLoader("nep_train", {0, 3}, 'NepTrainKit.core.io.nep:NepTrainResultData'))
register_result_loader(NepModelTypeLoader("nep_dipole", {1}, 'NepTrainKit.core.io.nep:NepDipoleResultData'))
register_result_loader(NepModelTypeLoader("nep_polar", {2}, 'NepTrainKit.core.io.nep:NepPolarizabilityResultData'))
register_result_loader(OtherLoader())
